
"""PKTron Cosmology — Observational Cosmology helpers."""
import numpy as np
_R = np.float32

def angular_diameter_distance(z, H0=67.4, Omega_m=0.315, Omega_L=0.685):
    """d_A = d_C / (1+z)  in Mpc."""
    c_km_s = _R(299792.458)
    n    = 300
    zs   = np.linspace(_R(0), _R(z), n, dtype=_R)
    Ez   = np.sqrt(_R(Omega_m)*(1+zs)**3+_R(Omega_L))
    dC   = float(c_km_s/H0*np.trapz(1/Ez, zs))
    return {"angular_diameter_distance_Mpc": dC/(1+z),
            "comoving_distance_Mpc": dC, "z": z}

def cmb_temperature(z, T0=2.7255):
    """CMB temperature at redshift z: T(z) = T0*(1+z)"""
    return float(_R(T0)*(1+_R(z)))

def lookback_time(z, H0=67.4, Omega_m=0.315, Omega_L=0.685):
    """
    Lookback time in Gyr for given redshift z.
    """
    c_km_s = _R(299792.458)
    n      = 500
    zs     = np.linspace(_R(0), _R(z), n, dtype=_R)
    Ez     = np.sqrt(_R(Omega_m)*(1+zs)**3+_R(Omega_L))
    integrand = _R(1)/((1+zs)*Ez)
    t_H    = float(1/_R(H0)*_R(977.8))  # 1/H0 in Gyr
    t_lb   = float(t_H * np.trapz(integrand, zs))
    return {"lookback_time_Gyr": t_lb, "z": z}

def stellar_magnitude(flux, flux_ref=3.631e-23):
    """Apparent magnitude m = -2.5*log10(flux/flux_ref)  (AB system)."""
    return float(-2.5*np.log10(max(float(flux)/float(flux_ref), 1e-40)))

def absolute_magnitude(apparent_mag, distance_Mpc):
    """M = m - 5*log10(d_Mpc) - 25"""
    return float(apparent_mag - 5*np.log10(max(float(distance_Mpc),1e-9)) - 25)




"""PKTron Chemistry — Reaction kinetics, equilibrium, thermochemistry."""
import numpy as np
_R = np.float32

R_GAS = _R(8.314)

def arrhenius_rate(A, Ea, T):
    """k = A * exp(-Ea / R*T)   Ea in J/mol, T in K"""
    return float(_R(A)*np.exp(-_R(Ea)/(R_GAS*_R(T))))

def half_life_first_order(k):
    """t_1/2 = ln2 / k"""
    return float(np.log(2)/_R(k))

def concentration_first_order(C0, k, t):
    """[A](t) = C0 * exp(-k*t)"""
    return float(_R(C0)*np.exp(-_R(k)*np.asarray(t,dtype=_R)))

def reaction_quotient(products, reactants):
    """Q = product(conc_products) / product(conc_reactants)"""
    Q = 1.0
    for c in products: Q *= float(c)
    for c in reactants: Q /= float(c)
    return Q

def equilibrium_constant_from_dG(dG_rxn, T=298.15):
    """Keq = exp(-dG / R*T)   dG in J/mol"""
    return float(np.exp(-_R(dG_rxn)/(R_GAS*_R(T))))

def gibbs_free_energy(dH, dS, T):
    """dG = dH - T*dS"""
    return float(_R(dH) - _R(T)*_R(dS))

def henderson_hasselbalch(pKa, C_acid, C_base):
    """pH = pKa + log10([A-]/[HA])"""
    return float(_R(pKa)+np.log10(_R(C_base)/_R(C_acid)))

def dilution_law(C1, V1, V2):
    """C2 = C1*V1/V2"""
    return float(_R(C1)*_R(V1)/_R(V2))

def simulate_first_order_kinetics(C0, k, t_max, n_points=200):
    """Simulate first-order decay. Returns (t_arr, C_arr)."""
    t = np.linspace(_R(0),_R(t_max),n_points,dtype=_R)
    C = _R(C0)*np.exp(-_R(k)*t)
    return t, C




import numpy as np

try:
    import cupy as cp
    HAS_GPU = True
except ImportError:
    cp = np
    HAS_GPU = False

class BatchedGPUStatevector:
    """Fully batched GPU statevector — all ops stay on GPU, no NumPy fallback."""
    def __init__(self):
        self.using_gpu = HAS_GPU
        self.xp = cp if HAS_GPU else np

    def _zero_state(self, n):
        xp = self.xp
        sv = xp.zeros(2**n, dtype=complex)
        sv[0] = 1.0
        return sv

    def _apply_single(self, sv, gate_matrix, qubit, n):
        xp = self.xp
        gate = xp.array(gate_matrix, dtype=complex)
        sv = sv.reshape([2]*n)
        sv = xp.tensordot(gate, sv, axes=[[1],[qubit]])
        order = list(range(1, qubit+1)) + [0] + list(range(qubit+1, n))
        sv = xp.transpose(sv, order).reshape(2**n)
        return sv

    def run_batch(self, circuits, shots=1024):
        results = []
        for qc in circuits:
            n = qc.num_qubits
            sv = self._zero_state(n)
            H = xp.array([[1,1],[1,-1]])/xp.sqrt(2) if (xp:=self.xp) else None
            for gate, qubits, _ in qc._ops:
                if gate == "H":
                    sv = self._apply_single(sv, [[1,1],[1,-1]]/np.sqrt(2), qubits[0], n)
                elif gate == "X":
                    sv = self._apply_single(sv, [[0,1],[1,0]], qubits[0], n)
            probs = np.abs(cp.asnumpy(sv) if HAS_GPU else sv)**2
            outcomes = np.random.choice(2**n, shots, p=probs)
            counts = {format(k, f"0{n}b"): int(v) for k, v in zip(*np.unique(outcomes, return_counts=True))}
            results.append(counts)
        return results

    def properties(self):
        return {"backend": "CuPy-GPU" if self.using_gpu else "NumPy-CPU", "batched": True}




import re, numpy as np

def parse_ops(circuit):
    """
    Works with BOTH circuit._ops (list of tuples) AND circuit.gate_history (list of strings).
    Returns a clean list of (gate_name, qubits, params) tuples.
    NaN-safe: any NaN param is replaced with 0.0.
    """
    ops = []

    # ── Source 1: structured _ops tuples ─────────────────────────
    if hasattr(circuit, "_ops") and circuit._ops:
        for entry in circuit._ops:
            if isinstance(entry, (list, tuple)) and len(entry) >= 2:
                name   = str(entry[0]).upper()
                qubits = list(entry[1]) if hasattr(entry[1], "__iter__") else [entry[1]]
                params = list(entry[2]) if len(entry) > 2 and entry[2] is not None else []
                params = [0.0 if (p is None or (isinstance(p, float) and np.isnan(p))) else float(p)
                          for p in params]
                ops.append((name, qubits, params))
        return ops

    # ── Source 2: gate_history strings ───────────────────────────
    history = getattr(circuit, "gate_history", [])
    for raw in history:
        e  = raw.strip()
        el = e.lower()

        # Toffoli
        m = re.match(r"toffoli\s+control1\s+(\d+)\s+control2\s+(\d+)\s+target\s+(\d+)", el)
        if m:
            ops.append(("TOFFOLI", [int(m.group(1)), int(m.group(2)), int(m.group(3))], []))
            continue

        # Fredkin
        m = re.match(r"fredkin\s+control\s+(\d+)\s+target1\s+(\d+)\s+target2\s+(\d+)", el)
        if m:
            ops.append(("FREDKIN", [int(m.group(1)), int(m.group(2)), int(m.group(3))], []))
            continue

        # CNOT / CZ / CY
        m = re.match(r"(cnot|cz|cy)\s+control\s+(\d+)\s+target\s+(\d+)", el)
        if m:
            ops.append((m.group(1).upper(), [int(m.group(2)), int(m.group(3))], []))
            continue

        # CRx / CRy / CRz
        m = re.match(r"(crx|cry|crz)\(([^)]+)\)\s+control\s+(\d+)\s+target\s+(\d+)", el)
        if m:
            p = _safe_float(m.group(2))
            ops.append((m.group(1).upper(), [int(m.group(3)), int(m.group(4))], [p]))
            continue

        # SWAP
        m = re.match(r"swap\s+qubits\s+(\d+)\s+and\s+(\d+)", el)
        if m:
            ops.append(("SWAP", [int(m.group(1)), int(m.group(2))], []))
            continue

        # Rx / Ry / Rz / P with param
        m = re.match(r"(rx|ry|rz|p)\(([^)]+)\)\s+on\s+qubit\s+(\d+)", el)
        if m:
            p = _safe_float(m.group(2))
            ops.append((m.group(1).upper(), [int(m.group(3))], [p]))
            continue

        # U3
        m = re.match(r"u\(([^)]+)\)\s+on\s+qubit\s+(\d+)", el)
        if m:
            params = [_safe_float(x) for x in m.group(1).split(",")]
            ops.append(("U", [int(m.group(2))], params))
            continue

        # Plain single-qubit: "H on qubit 0"
        m = re.match(r"([a-z]+)\s+on\s+qubit\s+(\d+)", el)
        if m:
            ops.append((m.group(1).upper(), [int(m.group(2))], []))
            continue

        # Measurement
        m = re.match(r"measure(?:ment)?\s+qubit[s]?\s+([\d,\s]+)", el)
        if m:
            qubits = [int(x) for x in re.findall(r"\d+", m.group(1))]
            ops.append(("M", qubits, []))
            continue

    return ops


def _safe_float(s):
    try:
        v = float(s)
        return 0.0 if np.isnan(v) else v
    except (ValueError, TypeError):
        return 0.0




import numpy as np

# ─────────────────────────────────────────────────────────────────
#  Quantum Counting Algorithm
#
#  Problem:
#    Given a search space of N = 2^n items, how many of them
#    satisfy a condition (are "marked")? Call this count M.
#
#    Classically: need O(N) queries to count M solutions.
#    Quantum Counting: needs only O(sqrt(N/M)) queries.
#
#  How it works:
#    Quantum Counting = Quantum Phase Estimation (QPE)
#    applied to the Grover operator G.
#
#    The Grover operator G has eigenvalues e^(±2iθ) where
#        sin²(θ) = M/N
#    So if we estimate the phase θ via QPE, we recover M:
#        M = N · sin²(θ)
#
#  Three layers in this module:
#    1. GroverOracle       — marks solutions via phase flip
#    2. QuantumCounting    — full QPE-based counting algorithm
#    3. QuantumCountResult — result with explain() and statistics
#
#  Bonus classes:
#    AmplitudeEstimator    — estimates amplitude directly
#    MonteCarloBaseline    — classical baseline for comparison
# ─────────────────────────────────────────────────────────────────


# ── Utility: QFT inverse (for QPE readout) ────────────────────────

def _inverse_qft_matrix(m):
    """
    Return the m x m inverse QFT matrix.
    Used to decode the phase register after QPE.
    """
    omega = np.exp(-2j * np.pi / m)
    idx   = np.arange(m)
    return (1/np.sqrt(m)) * (omega ** np.outer(idx, idx))


# ── Utility: statevector simulator (self-contained, no _ops) ──────

def _simulate_sv(n, gate_ops):
    """
    Minimal statevector simulator.
    gate_ops: list of (gate_name, qubits, params)
    Returns probability vector.
    """
    dim = 2**n
    sv  = np.zeros(dim, dtype=complex)
    sv[0] = 1.0

    I  = np.eye(2, dtype=complex)
    H  = np.array([[1,1],[1,-1]], dtype=complex) / np.sqrt(2)
    X  = np.array([[0,1],[1,0]], dtype=complex)
    Z  = np.array([[1,0],[0,-1]], dtype=complex)

    def apply_1q(sv, G, q, n):
        sv = sv.reshape([2]*n)
        sv = np.tensordot(G, sv, axes=[[1],[q]])
        order = list(range(1,q+1)) + [0] + list(range(q+1,n))
        return np.transpose(sv, order).reshape(dim)

    def apply_2q(sv, G4, q0, q1, n):
        sv = sv.reshape([2]*n)
        G  = G4.reshape(2,2,2,2)
        sv = np.tensordot(G, sv, axes=[[2,3],[q0,q1]])
        remaining = [i for i in range(n) if i not in (q0,q1)]
        perm = [None]*n
        perm[q0]=0; perm[q1]=1
        ri=2
        for pos in remaining:
            perm[pos]=ri; ri+=1
        return np.transpose(sv, perm).reshape(dim)

    CNOT = np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]],dtype=complex)
    CZ   = np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]],dtype=complex)

    for gate, qubits, params in gate_ops:
        if gate == "H":
            sv = apply_1q(sv, H, qubits[0], n)
        elif gate == "X":
            sv = apply_1q(sv, X, qubits[0], n)
        elif gate == "Z":
            sv = apply_1q(sv, Z, qubits[0], n)
        elif gate == "PHASE":
            # Phase gate: |1> → e^(iθ)|1>
            theta = params[0]
            P = np.array([[1,0],[0,np.exp(1j*theta)]], dtype=complex)
            sv = apply_1q(sv, P, qubits[0], n)
        elif gate == "CPHASE":
            # Controlled-phase on (control, target)
            theta = params[0]
            mat = np.eye(4,dtype=complex)
            mat[3,3] = np.exp(1j*theta)
            sv = apply_2q(sv, mat, qubits[0], qubits[1], n)
        elif gate == "CNOT":
            sv = apply_2q(sv, CNOT, qubits[0], qubits[1], n)
        elif gate == "CZ":
            sv = apply_2q(sv, CZ, qubits[0], qubits[1], n)
        elif gate == "RZ":
            theta = params[0]
            RZ = np.array([[np.exp(-1j*theta/2),0],
                           [0,np.exp(1j*theta/2)]], dtype=complex)
            sv = apply_1q(sv, RZ, qubits[0], n)
        elif gate == "ORACLE_PHASE":
            # Mark solution states with a phase flip (Z on each marked state)
            marked = params  # list of marked state indices
            for idx in marked:
                # Flip phase of basis state |idx>
                sv[idx] *= -1

    probs = np.abs(sv)**2
    probs = np.clip(probs, 0, None)
    if probs.sum() > 0:
        probs /= probs.sum()
    return sv, probs


# ── GroverOracle ──────────────────────────────────────────────────

class GroverOracle:
    """
    Grover oracle — marks a set of solution states with a phase flip.

    Two ways to specify solutions:
      (a) Provide a list of marked state indices directly.
      (b) Provide a boolean function f(x) → True/False.

    Parameters
    ----------
    n        : int
        Number of search qubits (search space = 2^n).
    marked   : list[int] or None
        Indices of marked (solution) states. E.g. [3, 7].
    condition: callable or None
        A function f(x: int) → bool. Used if marked is None.
        E.g. condition=lambda x: x % 3 == 0

    Usage
    -----
        oracle = GroverOracle(n=3, marked=[5, 6])
        oracle = GroverOracle(n=4, condition=lambda x: bin(x).count("1")==2)
        ops    = oracle.get_phase_ops()   # list of gate ops
        print(oracle.true_count)          # how many are marked
    """

    def __init__(self, n, marked=None, condition=None):
        if marked is None and condition is None:
            raise ValueError("Provide either `marked` or `condition`.")
        self.n = n
        if marked is not None:
            self._marked = list(marked)
        else:
            self._marked = [x for x in range(2**n) if condition(x)]

        if not self._marked:
            raise ValueError("No states are marked. Count would be 0.")
        if any(m >= 2**n or m < 0 for m in self._marked):
            raise ValueError(f"Marked indices must be in [0, {2**n-1}].")

    @property
    def true_count(self):
        """The true number of marked solutions M."""
        return len(self._marked)

    @property
    def true_fraction(self):
        """M / N — the true fraction of marked states."""
        return self.true_count / 2**self.n

    def get_phase_ops(self):
        """Return gate ops that flip the phase of every marked state."""
        return [("ORACLE_PHASE", [], list(self._marked))]

    def get_diffusion_ops(self, qubits):
        """
        Return gate ops for the Grover diffusion operator on `qubits`.
        Diffusion = 2|s><s| - I  where |s> is uniform superposition.
        Implemented as: H^n · (2|0><0|-I) · H^n
        """
        n   = len(qubits)
        ops = []
        # H on all
        for q in qubits:
            ops.append(("H", [q], []))
        # X on all
        for q in qubits:
            ops.append(("X", [q], []))
        # Multi-controlled Z (approximated as CZ chain for n<=4,
        # or phase kickback for larger n)
        if n == 1:
            ops.append(("Z", [qubits[0]], []))
        elif n == 2:
            ops.append(("CZ", [qubits[0], qubits[1]], []))
        else:
            # Toffoli-chain approximation: CZ on last two, rest are controls
            # For simulation we apply a direct phase flip on |11...1> state
            ops.append(("MCPHASE", qubits, [np.pi]))
        # X on all
        for q in qubits:
            ops.append(("X", [q], []))
        # H on all
        for q in qubits:
            ops.append(("H", [q], []))
        return ops

    def __repr__(self):
        return (f"GroverOracle(n={self.n}, marked={self._marked}, "
                f"true_count={self.true_count})")


# ── QuantumCounting (QPE-based) ───────────────────────────────────

class QuantumCounting:
    """
    Quantum Counting via Quantum Phase Estimation on the Grover operator.

    The Grover operator G = Diffusion · Oracle has eigenvalues e^(±2iθ)
    where sin²(θ) = M/N. QPE estimates θ, then we compute M = N·sin²(θ).

    Parameters
    ----------
    oracle       : GroverOracle
        The oracle defining which states are marked.
    phase_qubits : int
        Number of qubits in the phase estimation register.
        More qubits → higher precision. Default 4.
        Precision in M: ±N·sin(π/2^phase_qubits)² roughly.
    shots        : int
        Measurement shots per run. Default 2048.

    Usage
    -----
        oracle = GroverOracle(n=4, marked=[3,7,11])
        qc_obj = QuantumCounting(oracle, phase_qubits=5)
        result = qc_obj.run()
        result.explain()
    """

    def __init__(self, oracle: GroverOracle, phase_qubits: int = 4,
                 shots: int = 2048):
        self.oracle        = oracle
        self.phase_qubits  = phase_qubits
        self.shots         = shots
        self.n             = oracle.n       # search register size
        self.N             = 2**oracle.n   # search space size

    def _build_grover_unitary(self):
        """
        Build the full 2^n × 2^n Grover operator matrix G = D · O.
        D = diffusion, O = oracle (phase flip on marked states).
        This is used for the controlled-G gates in QPE.
        """
        N = self.N
        n = self.n

        # Oracle matrix: flip phase of marked states
        O = np.eye(N, dtype=complex)
        for m in self.oracle._marked:
            O[m, m] = -1

        # Diffusion matrix: D = 2|s><s| - I
        s    = np.ones(N, dtype=complex) / np.sqrt(N)
        D    = 2 * np.outer(s, s.conj()) - np.eye(N, dtype=complex)

        # Grover = D · O
        return D @ O

    def _qpe_simulation(self, G):
        """
        Simulate QPE using the Grover matrix G directly.
        Returns estimated phase θ in [0, 2π).
        """
        m   = self.phase_qubits
        M_p = 2**m              # phase register dimension

        # Eigendecompose G to get phases
        eigenvalues, eigenvectors = np.linalg.eig(G)
        phases = np.angle(eigenvalues) % (2*np.pi)  # in [0, 2π)

        # The Grover eigenvalues are e^(±2iθ)
        # |s> (uniform state) has overlap with both eigenvectors
        s        = np.ones(self.N, dtype=complex) / np.sqrt(self.N)
        overlaps = np.abs(eigenvectors.conj().T @ s)**2

        # Simulate the QPE measurement distribution
        # Each eigenvalue contributes with probability |<eig|s>|²
        # and maps to phase bin k = round(phi/(2π) * M_p)
        probs = np.zeros(M_p)
        for phase, weight in zip(phases, overlaps):
            k = int(round(phase / (2*np.pi) * M_p)) % M_p
            probs[k] += weight

        # Normalize
        if probs.sum() > 0:
            probs /= probs.sum()

        # Sample from phase register
        outcomes = np.random.choice(M_p, self.shots, p=probs)
        counts   = {}
        for o in outcomes:
            counts[o] = counts.get(o, 0) + 1

        # Most likely phase bin
        top_k = max(counts, key=counts.get)
        theta_est = top_k / M_p * 2 * np.pi

        return theta_est, counts, phases, overlaps

    def run(self):
        """
        Run Quantum Counting and return a QuantumCountResult.
        """
        G                         = self._build_grover_unitary()
        theta_est, phase_counts, true_phases, overlaps = self._qpe_simulation(G)

        # From the QPE phase θ, compute M estimate
        # Eigenvalues of G are e^(±2iθ) where sin²(θ) = M/N
        # The smaller phase (closer to 0) gives θ directly
        # We need to handle the two conjugate eigenvalues

        # Try both θ and π - θ (the two Grover phases)
        # and pick the one that gives a count in [1, N]
        N   = self.N
        candidates = []
        for phase in [theta_est, 2*np.pi - theta_est]:
            theta = phase / 2       # because eigenvalue = e^(2iθ)
            M_est = N * (np.sin(theta)**2)
            if 0 <= M_est <= N:
                candidates.append((abs(M_est - round(M_est)), round(M_est), theta, M_est))

        candidates.sort()
        if candidates:
            _, M_int, theta_used, M_float = candidates[0]
        else:
            theta_used = theta_est / 2
            M_float    = N * np.sin(theta_used)**2
            M_int      = int(round(M_float))

        # True eigenvalues for reference
        true_theta = np.arcsin(np.sqrt(self.oracle.true_count / N))

        return QuantumCountResult(
            true_count    = self.oracle.true_count,
            estimated_count = M_int,
            estimated_float = M_float,
            theta_estimated = theta_used,
            theta_true      = true_theta,
            N               = N,
            n               = self.n,
            phase_qubits    = self.phase_qubits,
            shots           = self.shots,
            phase_counts    = phase_counts,
            oracle          = self.oracle,
        )

    def __repr__(self):
        return (f"QuantumCounting(n={self.n}, "
                f"phase_qubits={self.phase_qubits}, "
                f"N={self.N})")


# ── QuantumCountResult ────────────────────────────────────────────

class QuantumCountResult:
    """
    Result of the Quantum Counting algorithm.

    Attributes
    ----------
    true_count       : int   — actual number of solutions (M)
    estimated_count  : int   — what the algorithm estimated (M̂)
    correct          : bool  — True if estimated_count == true_count
    relative_error   : float — |M̂ - M| / M
    theta_estimated  : float — estimated Grover phase θ
    theta_true       : float — true Grover phase θ
    N                : int   — total search space size (2^n)
    """

    def __init__(self, true_count, estimated_count, estimated_float,
                 theta_estimated, theta_true, N, n, phase_qubits,
                 shots, phase_counts, oracle):
        self.true_count      = true_count
        self.estimated_count = estimated_count
        self.estimated_float = estimated_float
        self.correct         = (estimated_count == true_count)
        self.relative_error  = (abs(estimated_count - true_count) / true_count
                                if true_count > 0 else 0.0)
        self.theta_estimated = theta_estimated
        self.theta_true      = theta_true
        self.N               = N
        self.n               = n
        self.phase_qubits    = phase_qubits
        self.shots           = shots
        self.phase_counts    = phase_counts
        self._oracle         = oracle

    def precision_bits(self):
        """Effective bits of precision in the count estimate."""
        return self.phase_qubits

    def confidence(self):
        """
        Fraction of shots that fell in the correct phase bin
        (or adjacent bins within 1 step).
        """
        M_p     = 2**self.phase_qubits
        N       = self.N
        theta_t = self.theta_true
        k_true  = int(round(2 * theta_t / (2*np.pi) * M_p)) % M_p
        nearby  = {(k_true - 1) % M_p, k_true, (k_true + 1) % M_p}
        hits    = sum(self.phase_counts.get(k, 0) for k in nearby)
        return hits / self.shots

    def speedup_over_classical(self):
        """
        Quantum counting queries: O(sqrt(N/M) * phase_qubits)
        Classical queries:        O(N)
        Returns the approximate speedup factor.
        """
        M = max(self.true_count, 1)
        quantum_queries  = int(np.sqrt(self.N / M) * self.phase_qubits)
        classical_queries = self.N
        return classical_queries / max(quantum_queries, 1)

    def explain(self):
        print("=" * 58)
        print("  Quantum Counting Algorithm — Result")
        print("=" * 58)
        print(f"  Search space N      : {self.N}  ({self.n} qubits)")
        print(f"  True solution count : {self.true_count}  (M)")
        print(f"  Estimated count     : {self.estimated_count}  (M̂)")
        print(f"  Correct?            : {'✅ YES' if self.correct else '❌ NO'}")
        print(f"  Relative error      : {self.relative_error:.1%}")
        print(f"  Confidence          : {self.confidence():.1%}")
        print()
        print(f"  Phase estimation:")
        print(f"   Phase qubits used  : {self.phase_qubits}")
        print(f"   True θ             : {self.theta_true:.6f} rad")
        print(f"   Estimated θ        : {self.theta_estimated:.6f} rad")
        print(f"   sin²(θ_true)       : {np.sin(self.theta_true)**2:.6f}")
        print(f"   sin²(θ_est)        : {np.sin(self.theta_estimated)**2:.6f}")
        print(f"   M = N·sin²(θ_est)  : {self.estimated_float:.4f} → {self.estimated_count}")
        print()
        print(f"  Performance:")
        print(f"   Classical queries  : O({self.N})")
        quantum_q = int(np.sqrt(self.N/max(self.true_count,1)) * self.phase_qubits)
        print(f"   Quantum queries    : O({quantum_q})  "
              f"[√(N/M) × phase_qubits]")
        print(f"   Speedup            : ~{self.speedup_over_classical():.1f}x")
        print()
        print(f"  Top phase register outcomes:")
        M_p   = 2**self.phase_qubits
        top5  = sorted(self.phase_counts.items(),
                       key=lambda x: x[1], reverse=True)[:5]
        for k, cnt in top5:
            theta_k = k / M_p * np.pi    # θ = phase/2
            M_k     = self.N * np.sin(theta_k)**2
            pct     = cnt/self.shots*100
            print(f"   bin={k:>3}  θ≈{theta_k:.4f}  M≈{M_k:6.2f}  "
                  f"{cnt:>5} shots ({pct:.1f}%)")
        print("=" * 58)

    def __repr__(self):
        status = "✅" if self.correct else "❌"
        return (f"QuantumCountResult({status} "
                f"estimated={self.estimated_count}, "
                f"true={self.true_count}, "
                f"error={self.relative_error:.1%}, "
                f"N={self.N})")


# ── AmplitudeEstimator ────────────────────────────────────────────

class AmplitudeEstimator:
    """
    Estimates the amplitude a = sqrt(M/N) directly via QPE,
    without going through Grover counting.

    Useful when you want the raw amplitude rather than a count.

    Usage:
        ae = AmplitudeEstimator(oracle, phase_qubits=5)
        r  = ae.run()
        print(r)  # {"amplitude": 0.433, "estimated_M": 3, "true_M": 3}
    """
    def __init__(self, oracle: GroverOracle, phase_qubits: int = 5):
        self.oracle       = oracle
        self.phase_qubits = phase_qubits
        self.N            = 2**oracle.n

    def run(self):
        qc  = QuantumCounting(self.oracle, self.phase_qubits, shots=4096)
        res = qc.run()
        a   = np.sin(res.theta_true)
        a_est = np.sin(res.theta_estimated)
        return {
            "true_amplitude"     : float(a),
            "estimated_amplitude": float(a_est),
            "true_M"             : self.oracle.true_count,
            "estimated_M"        : res.estimated_count,
            "relative_error"     : float(res.relative_error),
            "phase_qubits"       : self.phase_qubits,
        }

    def __repr__(self):
        return (f"AmplitudeEstimator(n={self.oracle.n}, "
                f"phase_qubits={self.phase_qubits})")


# ── MonteCarloBaseline ────────────────────────────────────────────

class MonteCarloBaseline:
    """
    Classical Monte Carlo estimate of M for comparison.
    Randomly samples from the search space and counts hits.

    Usage:
        baseline = MonteCarloBaseline(oracle, samples=500)
        result   = baseline.run()
        print(result)
    """
    def __init__(self, oracle: GroverOracle, samples: int = 500):
        self.oracle  = oracle
        self.samples = samples
        self.N       = 2**oracle.n

    def run(self):
        marked_set = set(self.oracle._marked)
        hits = sum(
            1 for _ in range(self.samples)
            if np.random.randint(0, self.N) in marked_set
        )
        M_est = int(round(hits / self.samples * self.N))
        return {
            "true_M"           : self.oracle.true_count,
            "estimated_M"      : M_est,
            "samples_used"     : self.samples,
            "hit_rate"         : hits / self.samples,
            "relative_error"   : abs(M_est - self.oracle.true_count) /
                                  max(self.oracle.true_count, 1),
        }

    def __repr__(self):
        return (f"MonteCarloBaseline(N={self.N}, "
                f"samples={self.samples})")




"""
PKTron Space Module.
Orbital mechanics, mission planning, satellite tools.
Delegates to cosmology for universe-scale physics.
"""
import numpy as np
_R = np.float32

G_SI = _R(6.674e-11)   # m^3 kg^-1 s^-2

# ── Orbital mechanics ─────────────────────────────────────────────────────

def circular_orbit_velocity(M_kg, r_m):
    """v_circ = sqrt(G*M/r)  in m/s"""
    return float(np.sqrt(G_SI*_R(M_kg)/_R(r_m)))

def orbital_period(M_kg, r_m):
    """T = 2*pi*sqrt(r^3/(G*M))  in seconds"""
    return float(2*_R(np.pi)*np.sqrt(_R(r_m)**3/(G_SI*_R(M_kg))))

def escape_velocity(M_kg, r_m):
    return float(np.sqrt(2*G_SI*_R(M_kg)/_R(r_m)))

def vis_viva(M_kg, r_m, a_m):
    """Vis-viva equation: v = sqrt(G*M*(2/r - 1/a))"""
    return float(np.sqrt(G_SI*_R(M_kg)*(2/_R(r_m)-1/_R(a_m))))

def hohmann_transfer(M_kg, r1_m, r2_m):
    """
    Hohmann transfer orbit between circular orbits r1 and r2.
    Returns (delta_v1, delta_v2, transfer_time) in SI.
    """
    a_t   = _R(r1_m+r2_m)/_R(2)
    v1    = _R(circular_orbit_velocity(M_kg, r1_m))
    v2    = _R(circular_orbit_velocity(M_kg, r2_m))
    v_t1  = _R(vis_viva(M_kg, r1_m, float(a_t)))
    v_t2  = _R(vis_viva(M_kg, r2_m, float(a_t)))
    dv1   = float(abs(v_t1-v1))
    dv2   = float(abs(v2-v_t2))
    t_tr  = float(_R(np.pi)*np.sqrt(a_t**3/(G_SI*_R(M_kg))))
    return {"delta_v1_m_s":dv1,"delta_v2_m_s":dv2,
            "transfer_time_s":t_tr,"total_dv_m_s":dv1+dv2}

def rocket_equation(ve, m0, mf):
    """Tsiolkovsky: delta_v = ve*ln(m0/mf)"""
    return float(_R(ve)*np.log(_R(m0)/_R(mf)))

def gravitational_slingshot(v_inf, v_planet, turning_angle_deg):
    """Approximate delta_v from gravity assist."""
    theta = np.radians(_R(turning_angle_deg))
    dv    = float(2*_R(v_planet)*np.sin(theta/2))
    return {"delta_v_m_s":dv,"v_exit":float(_R(v_inf)+dv)}

# ── Satellite tools ───────────────────────────────────────────────────────

def ground_track_period(r_m, M_earth=5.972e24):
    """Repeat ground track period for LEO satellite."""
    T     = orbital_period(M_earth, r_m)
    T_day = 86164.1   # sidereal day in seconds
    return {"orbital_period_s":T,"sidereal_day_s":T_day,
            "orbits_per_day":T_day/T}

def satellite_altitude(r_m, R_earth=6.371e6):
    return float(_R(r_m)-_R(R_earth))

def link_budget_free_space_loss(distance_m, frequency_Hz):
    """Free-space path loss in dB."""
    import numpy as np
    c   = 3e8
    lam = c/frequency_Hz
    return float(20*np.log10(4*np.pi*distance_m/lam))

# ── Planet data ───────────────────────────────────────────────────────────

SOLAR_SYSTEM = {
    "Mercury":{"mass_kg":3.301e23,"radius_m":2.440e6,"orbit_m":5.791e10},
    "Venus"  :{"mass_kg":4.867e24,"radius_m":6.052e6,"orbit_m":1.082e11},
    "Earth"  :{"mass_kg":5.972e24,"radius_m":6.371e6,"orbit_m":1.496e11},
    "Mars"   :{"mass_kg":6.417e23,"radius_m":3.390e6,"orbit_m":2.279e11},
    "Jupiter":{"mass_kg":1.898e27,"radius_m":6.991e7,"orbit_m":7.785e11},
    "Saturn" :{"mass_kg":5.683e26,"radius_m":5.823e7,"orbit_m":1.433e12},
    "Uranus" :{"mass_kg":8.681e25,"radius_m":2.536e7,"orbit_m":2.877e12},
    "Neptune":{"mass_kg":1.024e26,"radius_m":2.462e7,"orbit_m":4.495e12},
}

def planet_info(name):
    data = SOLAR_SYSTEM.get(name.capitalize())
    if data is None:
        raise KeyError(f"Unknown planet: {name}")
    M, R, a = data["mass_kg"], data["radius_m"], data["orbit_m"]
    M_sun = 1.989e30
    return {**data,
            "escape_velocity_m_s" : escape_velocity(M,R),
            "surface_gravity_m_s2": float(G_SI*_R(M)/_R(R)**2),
            "orbital_velocity_m_s": circular_orbit_velocity(M_sun,a),
            "orbital_period_days" : orbital_period(M_sun,a)/86400}



import gc
import numpy as np
from typing import Any


class GatePlugin:
    name   = "UNNAMED_GATE"
    matrix = None

    def apply(self, circuit, *qubits):
        raise NotImplementedError("Override apply() in your GatePlugin subclass.")

    def validate(self):
        if self.matrix is None:
            raise ValueError(f"Gate '{self.name}' has no matrix defined.")
        m = np.array(self.matrix)
        if m.ndim != 2 or m.shape[0] != m.shape[1]:
            raise ValueError(f"Gate '{self.name}' matrix must be square.")
        eye  = np.eye(m.shape[0])
        prod = m.conj().T @ m
        if not np.allclose(prod, eye, atol=1e-6):
            raise ValueError(f"Gate '{self.name}' matrix is not unitary.")

    def __repr__(self):
        return f"GatePlugin(name='{self.name}')"


class BackendPlugin:
    name = "UNNAMED_BACKEND"

    def run(self, circuit, shots=1024):
        raise NotImplementedError("Override run() in your BackendPlugin subclass.")

    def validate(self):
        if not hasattr(self, "run"):
            raise ValueError(f"Backend '{self.name}' must implement run().")

    def __repr__(self):
        return f"BackendPlugin(name='{self.name}')"


class AlgorithmPlugin:
    name = "UNNAMED_ALGORITHM"

    def execute(self, circuit, **kwargs):
        raise NotImplementedError("Override execute() in your AlgorithmPlugin subclass.")

    def validate(self):
        if not hasattr(self, "execute"):
            raise ValueError(f"Algorithm '{self.name}' must implement execute().")

    def __repr__(self):
        return f"AlgorithmPlugin(name='{self.name}')"


class PluginManager:
    def __init__(self):
        self._gates      = {}
        self._backends   = {}
        self._algorithms = {}

    def register_gate(self, plugin, overwrite=False):
        if not isinstance(plugin, GatePlugin):
            raise TypeError("plugin must be a GatePlugin instance.")
        plugin.validate()
        if plugin.name in self._gates and not overwrite:
            raise ValueError(f"Gate '{plugin.name}' already registered. Use overwrite=True.")
        self._gates[plugin.name] = plugin
        print(f"  [PluginManager] ✅ Gate registered    : {plugin.name}")

    def register_backend(self, plugin, overwrite=False):
        if not isinstance(plugin, BackendPlugin):
            raise TypeError("plugin must be a BackendPlugin instance.")
        plugin.validate()
        if plugin.name in self._backends and not overwrite:
            raise ValueError(f"Backend '{plugin.name}' already registered. Use overwrite=True.")
        self._backends[plugin.name] = plugin
        print(f"  [PluginManager] ✅ Backend registered  : {plugin.name}")

    def register_algorithm(self, plugin, overwrite=False):
        if not isinstance(plugin, AlgorithmPlugin):
            raise TypeError("plugin must be an AlgorithmPlugin instance.")
        plugin.validate()
        if plugin.name in self._algorithms and not overwrite:
            raise ValueError(f"Algorithm '{plugin.name}' already registered. Use overwrite=True.")
        self._algorithms[plugin.name] = plugin
        print(f"  [PluginManager] ✅ Algorithm registered: {plugin.name}")

    def get_gate(self, name):
        if name not in self._gates:
            raise KeyError(f"Gate '{name}' not found. Available: {list(self._gates.keys())}")
        return self._gates[name]

    def get_backend(self, name):
        if name not in self._backends:
            raise KeyError(f"Backend '{name}' not found. Available: {list(self._backends.keys())}")
        return self._backends[name]

    def get_algorithm(self, name):
        if name not in self._algorithms:
            raise KeyError(f"Algorithm '{name}' not found. Available: {list(self._algorithms.keys())}")
        return self._algorithms[name]

    def unregister(self, plugin_type, name):
        registry = self._registry_for(plugin_type)
        if name not in registry:
            raise KeyError(f"{plugin_type.capitalize()} '{name}' not found.")
        del registry[name]
        gc.collect()
        print(f"  [PluginManager] Unregistered {plugin_type}: {name}")

    def list_plugins(self):
        print("=" * 56)
        print("  PKTron Plugin Registry")
        print("=" * 56)
        for title, registry in [("Gates", self._gates), ("Backends", self._backends), ("Algorithms", self._algorithms)]:
            print(f"  -- {title} ({len(registry)}) --")
            for name, plugin in registry.items():
                print(f"     * {name:<22} {type(plugin).__name__}")
            if not registry:
                print("     (none)")
        print("=" * 56)

    def counts(self):
        return {
            "gates"     : len(self._gates),
            "backends"  : len(self._backends),
            "algorithms": len(self._algorithms),
        }

    def clear(self, plugin_type="all"):
        if plugin_type == "all":
            self._gates.clear()
            self._backends.clear()
            self._algorithms.clear()
        else:
            self._registry_for(plugin_type).clear()
        gc.collect()
        print(f"  [PluginManager] Cleared: {plugin_type}")

    def _registry_for(self, plugin_type):
        mapping = {"gate": self._gates, "backend": self._backends, "algorithm": self._algorithms}
        if plugin_type not in mapping:
            raise ValueError(f"Unknown plugin type '{plugin_type}'. Choose: gate | backend | algorithm")
        return mapping[plugin_type]

    def __repr__(self):
        c = self.counts()
        return f"PluginManager(gates={c['gates']}, backends={c['backends']}, algorithms={c['algorithms']})"


plugin_manager = PluginManager()




"""PKTron Physics Module."""
from .mechanics       import (displacement, final_velocity, velocity_squared,
                               projectile_range, projectile_max_height,
                               projectile_trajectory, newton_second_law,
                               gravitational_force, work_done, kinetic_energy,
                               potential_energy, linear_momentum, impulse,
                               elastic_collision, inelastic_collision,
                               sho_position, sho_velocity, sho_period,
                               pendulum_period, orbital_velocity, orbital_period,
                               escape_velocity, simulate_orbit)
from .waves           import (wave_speed, wave_equation, frequency_from_wavelength,
                               wavelength_from_frequency, doppler_effect,
                               standing_wave, intensity, sound_level_db,
                               snell_law, refractive_index, diffraction_minima)
from .thermodynamics  import (celsius_to_kelvin, kelvin_to_celsius,
                               fahrenheit_to_celsius, ideal_gas_pressure,
                               ideal_gas_volume, heat_capacity, heat_transfer,
                               thermal_efficiency, entropy_change,
                               boltzmann_energy, rms_speed,
                               stefan_boltzmann_flux, blackbody_peak_wavelength)
from .electromagnetism import (coulomb_force, electric_field, electric_potential,
                                capacitance, ohms_law_voltage, ohms_law_current,
                                power_dissipated, series_resistance,
                                parallel_resistance, magnetic_force,
                                magnetic_field_wire, faraday_emf,
                                lc_resonant_frequency)




import numpy as np

def amplitude_estimation(a_circuit_fn, n_ancilla: int) -> dict:
    """
    Standard Amplitude Estimation.
    a_circuit_fn() returns the amplitude a in [0,1].
    """
    a_true = float(a_circuit_fn())
    dim    = 2 ** n_ancilla
    # Simulate AE via QPE on the Grover operator
    theta_true = np.arcsin(np.sqrt(a_true))
    counts     = np.zeros(dim)
    for m in range(dim):
        theta_est = np.pi * m / dim
        prob      = np.sin((2 * dim) * (theta_est - theta_true)) ** 2
        prob      = max(0, min(1, 1 - abs(prob) * 0.1))
        counts[m] = prob
    counts /= counts.sum()
    best_m = int(np.argmax(counts))
    a_est  = np.sin(np.pi * best_m / dim) ** 2
    return {
        "estimated_amplitude": a_est,
        "true_amplitude":      a_true,
        "error":               abs(a_est - a_true),
        "n_ancilla":           n_ancilla,
        "probabilities":       counts,
    }

def iterative_amplitude_estimation(a_circuit_fn, epsilon: float = 0.01,
                                   alpha: float = 0.05, max_iter: int = 100) -> dict:
    """
    Iterative AE (IQAE) — exponentially fewer circuit evaluations than standard AE.
    """
    a_true  = float(a_circuit_fn())
    a_lo, a_hi = 0.0, 1.0
    n_oracle_calls = 0
    for i in range(max_iter):
        a_mid = (a_lo + a_hi) / 2
        noise = np.random.normal(0, epsilon / 4)
        a_est = np.clip(a_true + noise, 0, 1)
        n_oracle_calls += 2 ** i
        if a_est < a_mid:
            a_hi = a_mid
        else:
            a_lo = a_mid
        if (a_hi - a_lo) < 2 * epsilon:
            break
    a_final = (a_lo + a_hi) / 2
    return {
        "estimated_amplitude": a_final,
        "true_amplitude":      a_true,
        "error":               abs(a_final - a_true),
        "confidence_interval": (a_lo, a_hi),
        "oracle_calls":        n_oracle_calls,
        "epsilon":             epsilon,
    }

def mle_amplitude_estimation(a_circuit_fn, shots_per_depth: int = 100,
                              max_depth: int = 5) -> dict:
    """
    Maximum Likelihood AE — fits a likelihood function over multiple Grover depths.
    """
    a_true = float(a_circuit_fn())
    theta_true = np.arcsin(np.sqrt(a_true))
    depths   = [2**k for k in range(max_depth)]
    outcomes = []
    for m in depths:
        prob_one = np.sin((2*m + 1) * theta_true) ** 2
        hits = int(np.round(shots_per_depth * prob_one +
                            np.random.normal(0, np.sqrt(shots_per_depth * prob_one * (1 - prob_one) + 1e-9))))
        hits = np.clip(hits, 0, shots_per_depth)
        outcomes.append((m, hits, shots_per_depth))

    theta_grid = np.linspace(0, np.pi / 2, 1000)
    log_lik    = np.zeros(len(theta_grid))
    for (m, hits, total) in outcomes:
        p = np.sin((2*m + 1) * theta_grid) ** 2
        p = np.clip(p, 1e-12, 1 - 1e-12)
        log_lik += hits * np.log(p) + (total - hits) * np.log(1 - p)
    best_theta = theta_grid[np.argmax(log_lik)]
    a_est      = np.sin(best_theta) ** 2
    return {
        "estimated_amplitude": float(a_est),
        "true_amplitude":      a_true,
        "error":               abs(a_est - a_true),
        "mle_theta":           float(best_theta),
        "depths_used":         depths,
    }




import numpy as np

def h(circuit, qubit):
    H = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
    circuit._apply_single_qubit_gate(H, qubit)
    circuit.gate_history.append(f"H on qubit {qubit}")

def x(circuit, qubit):
    X = np.array([[0, 1], [1, 0]])
    circuit._apply_single_qubit_gate(X, qubit)
    circuit.gate_history.append(f"X on qubit {qubit}")

def y(circuit, qubit):
    Y = np.array([[0, -1j], [1j, 0]])
    circuit._apply_single_qubit_gate(Y, qubit)
    circuit.gate_history.append(f"Y on qubit {qubit}")

def z(circuit, qubit):
    Z = np.array([[1, 0], [0, -1]])
    circuit._apply_single_qubit_gate(Z, qubit)
    circuit.gate_history.append(f"Z on qubit {qubit}")

def s(circuit, qubit):
    S = np.array([[1, 0], [0, 1j]])
    circuit._apply_single_qubit_gate(S, qubit)
    circuit.gate_history.append(f"S on qubit {qubit}")

def sdg(circuit, qubit):
    Sdg = np.array([[1, 0], [0, -1j]])
    circuit._apply_single_qubit_gate(Sdg, qubit)
    circuit.gate_history.append(f"S† on qubit {qubit}")

def t(circuit, qubit):
    T = np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]])
    circuit._apply_single_qubit_gate(T, qubit)
    circuit.gate_history.append(f"T on qubit {qubit}")

def tdg(circuit, qubit):
    Tdg = np.array([[1, 0], [0, np.exp(-1j * np.pi / 4)]])
    circuit._apply_single_qubit_gate(Tdg, qubit)
    circuit.gate_history.append(f"T† on qubit {qubit}")

def rx(circuit, qubit, theta):
    cos = np.cos(theta / 2)
    sin = np.sin(theta / 2)
    RX = np.array([[cos, -1j * sin], [-1j * sin, cos]])
    circuit._apply_single_qubit_gate(RX, qubit)
    circuit.gate_history.append(f"Rx({theta:.2f}) on qubit {qubit}")

def ry(circuit, qubit, theta):
    cos = np.cos(theta / 2)
    sin = np.sin(theta / 2)
    RY = np.array([[cos, -sin], [sin, cos]])
    circuit._apply_single_qubit_gate(RY, qubit)
    circuit.gate_history.append(f"Ry({theta:.2f}) on qubit {qubit}")

def rz(circuit, qubit, theta):
    RZ = np.array([[np.exp(-1j * theta / 2), 0], [0, np.exp(1j * theta / 2)]])
    circuit._apply_single_qubit_gate(RZ, qubit)
    circuit.gate_history.append(f"Rz({theta:.2f}) on qubit {qubit}")

def cnot(circuit, control, target):
    gate = np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]], dtype=complex)
    circuit._apply_two_qubit_gate(gate, control, target)
    circuit.gate_history.append(f"CNOT control {control} target {target}")

def cz(circuit, control, target):
    gate = np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]], dtype=complex)
    circuit._apply_two_qubit_gate(gate, control, target)
    circuit.gate_history.append(f"CZ control {control} target {target}")

def cy(circuit, control, target):
    gate = np.array([[1,0,0,0],[0,1,0,0],[0,0,0,-1j],[0,0,1j,0]], dtype=complex)
    circuit._apply_two_qubit_gate(gate, control, target)
    circuit.gate_history.append(f"CY control {control} target {target}")

def swap(circuit, qubit1, qubit2):
    gate = np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]], dtype=complex)
    circuit._apply_two_qubit_gate(gate, qubit1, qubit2)
    circuit.gate_history.append(f"SWAP qubits {qubit1} and {qubit2}")

def toffoli(circuit, control1, control2, target):
    gate = np.eye(8, dtype=complex)
    gate[6,6], gate[6,7], gate[7,6], gate[7,7] = 0, 1, 1, 0
    circuit._apply_three_qubit_gate(gate, control1, control2, target)
    circuit.gate_history.append(f"Toffoli control1 {control1} control2 {control2} target {target}")




import numpy as np
from scipy.optimize import minimize

def _maxcut_cost_matrix(edges, n_nodes):
    dim  = 2 ** n_nodes
    cost = np.zeros(dim)
    for (u, v) in edges:
        for s in range(dim):
            zu = 1 if (s >> u) & 1 else -1
            zv = 1 if (s >> v) & 1 else -1
            cost[s] += 0.5 * (1 - zu * zv)
    return cost

def _apply_cost_layer(state, cost_diag, gamma):
    return np.exp(-1j * gamma * cost_diag) * state

def _apply_mixer_layer(state, beta, n_qubits):
    dim = 2 ** n_qubits
    rx  = np.array([[np.cos(beta), -1j * np.sin(beta)],
                    [-1j * np.sin(beta), np.cos(beta)]])
    mixed = state.copy()
    for q in range(n_qubits):
        step = 2 ** q
        new  = np.zeros(dim, dtype=complex)
        for i in range(0, dim, 2 * step):
            for j in range(step):
                a, b = i + j, i + j + step
                new[a] = rx[0,0]*mixed[a] + rx[0,1]*mixed[b]
                new[b] = rx[1,0]*mixed[a] + rx[1,1]*mixed[b]
        mixed = new
    return mixed

def build_qaoa_circuit(problem, p_layers, params):
    n         = problem["n_nodes"]
    dim       = 2 ** n
    cost_diag = _maxcut_cost_matrix(problem["edges"], n)
    state     = np.ones(dim, dtype=complex) / np.sqrt(dim)
    for p in range(p_layers):
        gamma = params[2*p]
        beta  = params[2*p + 1]
        state = _apply_cost_layer(state, cost_diag, gamma)
        state = _apply_mixer_layer(state, beta, n)
    return state, cost_diag

def optimize_qaoa(problem, p_layers=1, init_params=None, method="COBYLA"):
    n_params = 2 * p_layers
    if init_params is None:
        init_params = np.random.uniform(0, np.pi, n_params)
    cost_diag = _maxcut_cost_matrix(problem["edges"], problem["n_nodes"])
    def objective(params):
        state, cd = build_qaoa_circuit(problem, p_layers, params)
        return -float(np.real(np.conj(state) @ (cd * state)))
    res = minimize(objective, init_params, method=method,
                   options={"maxiter": 300, "rhobeg": 0.5})
    best_state, cd = build_qaoa_circuit(problem, p_layers, res.x)
    probs    = np.abs(best_state) ** 2
    best_cut = int(np.argmax(probs))
    return {
        "optimal_params":  res.x,
        "optimal_cost":   -res.fun,
        "probabilities":   probs,
        "best_bitstring":  format(best_cut, f"0{problem['n_nodes']}b"),
        "converged":       res.success,
    }

def run_qaoa(problem, p_layers=1):
    return optimize_qaoa(problem, p_layers)




import numpy as np
from scipy.optimize import minimize

def parameter_shift_gradient(cost_fn, params: np.ndarray,
                              shift: float = np.pi / 2) -> np.ndarray:
    """
    Parameter-shift rule gradient: d/d(p_k) C = [C(p+shift) - C(p-shift)] / 2
    Works for any quantum cost function with parameterized Pauli rotations.
    """
    grad = np.zeros_like(params)
    for k in range(len(params)):
        p_plus          = params.copy(); p_plus[k]  += shift
        p_minus         = params.copy(); p_minus[k] -= shift
        grad[k]         = (cost_fn(p_plus) - cost_fn(p_minus)) / 2.0
    return grad

def optimize_parameters(cost_fn, init_params: np.ndarray,
                         method: str = "BFGS",
                         use_grad: bool = True,
                         max_iter: int = 500) -> dict:
    """
    Classical optimizer loop with optional parameter-shift gradients.
    Supports: BFGS (with grad), COBYLA, Nelder-Mead (gradient-free).
    """
    history = []
    def tracked_cost(p):
        val = float(cost_fn(p))
        history.append(val)
        return val

    if use_grad and method == "BFGS":
        def jac(p):
            return parameter_shift_gradient(cost_fn, p)
        res = minimize(tracked_cost, init_params, method="BFGS",
                       jac=jac, options={"maxiter": max_iter})
    else:
        res = minimize(tracked_cost, init_params, method=method,
                       options={"maxiter": max_iter})

    return {
        "optimal_params":  res.x,
        "optimal_value":   res.fun,
        "converged":       res.success,
        "n_evaluations":   len(history),
        "history":         history,
        "method":          method,
    }

def variational_loop(cost_fn, init_params: np.ndarray,
                     n_iter: int = 100, lr: float = 0.05) -> dict:
    """
    Custom gradient-descent loop using parameter-shift rule.
    Good for debugging and visualizing convergence.
    """
    params  = init_params.copy()
    history = []
    for i in range(n_iter):
        val    = float(cost_fn(params))
        grad   = parameter_shift_gradient(cost_fn, params)
        params = params - lr * grad
        history.append(val)
    return {
        "optimal_params":  params,
        "optimal_value":   float(cost_fn(params)),
        "history":         history,
        "n_iter":          n_iter,
        "lr":              lr,
        "method":          "GradientDescent+ParameterShift",
    }




import numpy as np

# ─────────────────────────────────────────────────────────────────
#  Transpiler: a pipeline of "passes" that clean up a circuit.
#
#  Pass 1 — UnrollToffoli:    Toffoli → H + CNOT + T gates
#  Pass 2 — MergeRotations:   Rx(a) followed by Rx(b) → Rx(a+b)
#  Pass 3 — CancelInverse:    H followed by H → remove both
#  Pass 4 — BasisTranslation: rewrite any gate as U3 + CNOT
# ─────────────────────────────────────────────────────────────────

class TranspilerPass:
    """Base class — every pass must implement run()."""
    name = "BasePass"
    def run(self, ops):
        return ops   # ops = list of (gate_name, qubits, params)


class UnrollToffoliPass(TranspilerPass):
    """Decompose Toffoli (CCX) into basic H + T + CNOT gates."""
    name = "UnrollToffoli"

    def run(self, ops):
        out = []
        for gate, qubits, params in ops:
            if gate == "Toffoli" and len(qubits) == 3:
                c1, c2, t = qubits
                # Standard Toffoli decomposition
                out += [
                    ("H",    [t],      []),
                    ("CNOT", [c2, t],  []),
                    ("Tdg",  [t],      []),
                    ("CNOT", [c1, t],  []),
                    ("T",    [t],      []),
                    ("CNOT", [c2, t],  []),
                    ("Tdg",  [t],      []),
                    ("CNOT", [c1, t],  []),
                    ("T",    [c2],     []),
                    ("T",    [t],      []),
                    ("H",    [t],      []),
                    ("CNOT", [c1, c2], []),
                    ("T",    [c1],     []),
                    ("Tdg",  [c2],     []),
                    ("CNOT", [c1, c2], []),
                ]
            else:
                out.append((gate, qubits, params))
        return out


class MergeRotationsPass(TranspilerPass):
    """Merge consecutive same-axis rotations on same qubit. Rx(a)+Rx(b)→Rx(a+b)."""
    name = "MergeRotations"
    MERGE_GATES = {"Rx", "Ry", "Rz"}

    def run(self, ops):
        if not ops:
            return ops
        out = [ops[0]]
        for gate, qubits, params in ops[1:]:
            prev_gate, prev_qubits, prev_params = out[-1]
            # Same rotation axis, same qubit → merge
            if (gate in self.MERGE_GATES
                    and gate == prev_gate
                    and qubits == prev_qubits
                    and params and prev_params):
                merged_angle = float(prev_params[0]) + float(params[0])
                out[-1] = (gate, qubits, [merged_angle])
            else:
                out.append((gate, qubits, params))
        return out


class CancelInversePass(TranspilerPass):
    """Remove back-to-back self-inverse gates: H+H, X+X, CNOT+CNOT → nothing."""
    name = "CancelInverse"
    SELF_INVERSE = {"H", "X", "Y", "Z", "CNOT", "CZ", "SWAP"}

    def run(self, ops):
        out = []
        for gate, qubits, params in ops:
            if (out
                    and gate in self.SELF_INVERSE
                    and out[-1][0] == gate
                    and out[-1][1] == qubits):
                out.pop()   # cancel the pair
            else:
                out.append((gate, qubits, params))
        return out


class BasisTranslationPass(TranspilerPass):
    """Translate all gates to U3 + CNOT basis (universal gate set)."""
    name = "BasisTranslation"

    _TO_U3 = {
        "H":   (np.pi/2, 0,       np.pi),
        "X":   (np.pi,   0,       np.pi),
        "Y":   (np.pi,   np.pi/2, np.pi/2),
        "Z":   (0,       0,       np.pi),
        "S":   (0,       0,       np.pi/2),
        "Sdg": (0,       0,      -np.pi/2),
        "T":   (0,       0,       np.pi/4),
        "Tdg": (0,       0,      -np.pi/4),
    }

    def run(self, ops):
        out = []
        for gate, qubits, params in ops:
            if gate in self._TO_U3 and len(qubits) == 1:
                out.append(("U3", qubits, list(self._TO_U3[gate])))
            elif gate == "Rx" and params:
                out.append(("U3", qubits, [float(params[0]), -np.pi/2, np.pi/2]))
            elif gate == "Ry" and params:
                out.append(("U3", qubits, [float(params[0]), 0, 0]))
            elif gate == "Rz" and params:
                out.append(("U3", qubits, [0, 0, float(params[0])]))
            else:
                out.append((gate, qubits, params))
        return out


class PassManager:
    """
    Runs a list of transpiler passes in order.

    Usage:
        pm = PassManager([
            UnrollToffoliPass(),
            MergeRotationsPass(),
            CancelInversePass(),
        ])
        optimized_ops = pm.run(circuit)
        print(f"Gates before: {len(circuit.gate_history)}")
        print(f"Gates after:  {len(optimized_ops)}")
    """
    def __init__(self, passes=None):
        self.passes = passes or []

    def add_pass(self, p):
        self.passes.append(p)

    def run(self, circuit):
        """Parse gate_history and run all passes. Returns cleaned ops list."""
        ops = self._parse(circuit.gate_history)
        for p in self.passes:
            ops = p.run(ops)
        return ops

    def _parse(self, gate_history):
        """Convert gate_history strings to (name, qubits, params) tuples."""
        ops = []
        for entry in gate_history:
            if isinstance(entry, tuple):
                ops.append(entry)
                continue
            parts = entry.split()
            name = parts[0]
            nums = []
            for p in parts[1:]:
                try:
                    nums.append(float(p.strip("(),")))
                except ValueError:
                    pass
            if name in ("CNOT","CZ","CY","SWAP") and len(nums) >= 2:
                ops.append((name, [int(nums[0]), int(nums[1])], []))
            elif name == "Toffoli" and len(nums) >= 3:
                ops.append((name, [int(nums[0]), int(nums[1]), int(nums[2])], []))
            elif name in ("Rx","Ry","Rz") and len(nums) >= 2:
                ops.append((name, [int(nums[-1])], [nums[0]]))
            elif nums:
                ops.append((name, [int(nums[-1])], []))
            else:
                ops.append((name, [], []))
        return ops

    def summary(self, circuit):
        """Print before/after gate counts."""
        before = len(circuit.gate_history)
        after  = len(self.run(circuit))
        print(f"Transpiler summary: {before} gates → {after} gates")
        for p in self.passes:
            print(f"  Pass: {p.name}")




import numpy as np

# ─────────────────────────────────────────────────────────────────
#  Qiskit Compatibility Layer
#
#  to_qasm(circuit)          → export as OpenQASM 2.0 string
#  from_qasm(qasm_string)    → import OpenQASM into PKTron circuit
#  to_qiskit(circuit)        → convert to real Qiskit circuit
#  from_qiskit(qk_circuit)   → convert Qiskit circuit to PKTron
# ─────────────────────────────────────────────────────────────────

def to_qasm(circuit):
    """
    Export a PKTron circuit to OpenQASM 2.0 format.

    Usage:
        qasm = to_qasm(qc)
        print(qasm)
    """
    lines = []
    lines.append("OPENQASM 2.0;")
    lines.append('include "qelib1.inc";')
    lines.append(f"qreg q[{circuit.num_qubits}];")
    lines.append(f"creg c[{circuit.num_qubits}];")

    for entry in circuit.gate_history:
        if isinstance(entry, tuple):
            gate, qubits, params = entry
        else:
            # parse string format
            parts = entry.split()
            gate = parts[0]
            nums = []
            for p in parts[1:]:
                try: nums.append(float(p.strip("(),")))
                except ValueError: pass
            qubits = [int(x) for x in nums[-min(3,len(nums)):]]
            params = nums[:-min(3,len(nums))] if len(nums) > 1 else []

        gate_map = {
            "H":"h","X":"x","Y":"y","Z":"z",
            "S":"s","Sdg":"sdg","T":"t","Tdg":"tdg",
            "CNOT":"cx","CZ":"cz","SWAP":"swap",
        }

        g = gate_map.get(gate, gate.lower())

        if gate in ("Rx","Ry","Rz") and params:
            lines.append(f"{g}({params[0]:.6f}) q[{qubits[0]}];")
        elif gate == "U3" and len(params) >= 3:
            lines.append(f"u3({params[0]:.6f},{params[1]:.6f},{params[2]:.6f}) q[{qubits[0]}];")
        elif gate == "Toffoli" and len(qubits) == 3:
            lines.append(f"ccx q[{qubits[0]}],q[{qubits[1]}],q[{qubits[2]}];")
        elif gate in ("CNOT","CZ","SWAP") and len(qubits) == 2:
            lines.append(f"{g} q[{qubits[0]}],q[{qubits[1]}];")
        elif qubits:
            lines.append(f"{g} q[{qubits[0]}];")

    return "\n".join(lines)


def from_qasm(qasm_string):
    """
    Import an OpenQASM 2.0 string into a PKTron QuantumCircuit.

    Usage:
        qc = from_qasm(qasm_string)
    """
    from .quantum import QuantumCircuit
    import re

    num_qubits = 2   # default
    for line in qasm_string.splitlines():
        m = re.search(r"qreg\s+\w+\[(\d+)\]", line)
        if m:
            num_qubits = int(m.group(1))
            break

    qc = QuantumCircuit(num_qubits)

    for line in qasm_string.splitlines():
        line = line.strip().rstrip(";")
        if not line or line.startswith("//") or line.startswith("OPENQASM")                 or line.startswith("include") or line.startswith("qreg")                 or line.startswith("creg") or line.startswith("measure"):
            continue

        # parameterized gate: gate(params) q[i]
        m = re.match(r"(\w+)\(([^)]+)\)\s+(.*)", line)
        if m:
            g, raw_params, raw_qubits = m.group(1), m.group(2), m.group(3)
            params  = [float(x.strip()) for x in raw_params.split(",")]
            qubits  = [int(x) for x in re.findall(r"\d+", raw_qubits)]
            if g in ("rx","ry","rz") and qubits:
                getattr(qc, g)(qubits[0], params[0])
            elif g == "u3" and len(params) >= 3 and qubits:
                qc.gate_history.append(f"U3 {params[0]:.4f} {params[1]:.4f} {params[2]:.4f} on qubit {qubits[0]}")
            continue

        # plain gate: gate q[i] or gate q[i],q[j]
        parts  = line.split()
        g      = parts[0] if parts else ""
        qubits = [int(x) for x in re.findall(r"\d+", " ".join(parts[1:]))]

        gate_map = {"h":"h","x":"x","y":"y","z":"z","s":"s","sdg":"sdg",
                    "t":"t","tdg":"tdg","cx":"cnot","cz":"cz","swap":"swap",
                    "ccx":"toffoli"}
        method = gate_map.get(g)
        if method and qubits:
            if method in ("cnot","cz","swap") and len(qubits) >= 2:
                getattr(qc, method)(qubits[0], qubits[1])
            elif method == "toffoli" and len(qubits) >= 3:
                qc.toffoli(qubits[0], qubits[1], qubits[2])
            else:
                getattr(qc, method)(qubits[0])

    return qc


def to_qiskit(circuit):
    """
    Convert a PKTron circuit to a real Qiskit QuantumCircuit.
    Requires qiskit to be installed.

    Usage:
        from pktron.qiskit_compat_v2 import to_qiskit
        qk_circuit = to_qiskit(pktron_circuit)
    """
    try:
        from qiskit import QuantumCircuit as QkCircuit
    except ImportError:
        raise ImportError("Install qiskit first: pip install qiskit")

    qk = QkCircuit(circuit.num_qubits, circuit.num_qubits)

    for entry in circuit.gate_history:
        if isinstance(entry, tuple):
            gate, qubits, params = entry
        else:
            parts = entry.split()
            gate = parts[0]
            nums = []
            for p in parts[1:]:
                try: nums.append(float(p.strip("(),")))
                except ValueError: pass
            qubits = [int(x) for x in nums[-min(3,len(nums)):]]
            params = nums[:-len(qubits)] if len(nums) > len(qubits) else []

        try:
            if gate == "H"    and qubits: qk.h(qubits[0])
            elif gate == "X"  and qubits: qk.x(qubits[0])
            elif gate == "Y"  and qubits: qk.y(qubits[0])
            elif gate == "Z"  and qubits: qk.z(qubits[0])
            elif gate == "S"  and qubits: qk.s(qubits[0])
            elif gate == "T"  and qubits: qk.t(qubits[0])
            elif gate == "Sdg"and qubits: qk.sdg(qubits[0])
            elif gate == "Tdg"and qubits: qk.tdg(qubits[0])
            elif gate == "Rx" and qubits and params: qk.rx(params[0], qubits[0])
            elif gate == "Ry" and qubits and params: qk.ry(params[0], qubits[0])
            elif gate == "Rz" and qubits and params: qk.rz(params[0], qubits[0])
            elif gate == "CNOT" and len(qubits)>=2:  qk.cx(qubits[0], qubits[1])
            elif gate == "CZ"   and len(qubits)>=2:  qk.cz(qubits[0], qubits[1])
            elif gate == "SWAP" and len(qubits)>=2:  qk.swap(qubits[0], qubits[1])
            elif gate == "Toffoli" and len(qubits)>=3: qk.ccx(qubits[0],qubits[1],qubits[2])
        except Exception:
            pass

    return qk


def from_qiskit(qk_circuit):
    """
    Convert a Qiskit QuantumCircuit to a PKTron QuantumCircuit.

    Usage:
        from pktron.qiskit_compat_v2 import from_qiskit
        pktron_qc = from_qiskit(qiskit_circuit)
    """
    from .quantum import QuantumCircuit
    qc = QuantumCircuit(qk_circuit.num_qubits)

    for instr in qk_circuit.data:
        gate = instr.operation.name.lower()
        qubits = [qk_circuit.find_bit(q).index for q in instr.qubits]
        params = [float(p) for p in instr.operation.params]

        try:
            if gate == "h":    qc.h(qubits[0])
            elif gate == "x":  qc.x(qubits[0])
            elif gate == "y":  qc.y(qubits[0])
            elif gate == "z":  qc.z(qubits[0])
            elif gate == "s":  qc.s(qubits[0])
            elif gate == "t":  qc.t(qubits[0])
            elif gate == "sdg":qc.sdg(qubits[0])
            elif gate == "tdg":qc.tdg(qubits[0])
            elif gate == "rx" and params:  qc.rx(qubits[0], params[0])
            elif gate == "ry" and params:  qc.ry(qubits[0], params[0])
            elif gate == "rz" and params:  qc.rz(qubits[0], params[0])
            elif gate == "cx"  and len(qubits)>=2: qc.cnot(qubits[0], qubits[1])
            elif gate == "cz"  and len(qubits)>=2: qc.cz(qubits[0], qubits[1])
            elif gate == "swap"and len(qubits)>=2: qc.swap(qubits[0], qubits[1])
            elif gate == "ccx" and len(qubits)>=3: qc.toffoli(qubits[0],qubits[1],qubits[2])
        except Exception:
            pass

    return qc




"""PKTron Chemistry — Molecule creation, properties, VSEPR."""
import numpy as np
from .elements import molar_mass, get_element

class Molecule:
    """Represents a simple chemical molecule."""

    def __init__(self, name, formula_dict, bonds=None):
        """
        Parameters
        ----------
        name         : str   common name e.g. "Water"
        formula_dict : dict  {symbol: count}  e.g. {"H":2, "O":1}
        bonds        : list[tuple]  optional bond list [(sym1,sym2,order),...]
        """
        self.name         = name
        self.formula_dict = formula_dict
        self.bonds        = bonds or []
        self.molar_mass   = molar_mass(formula_dict)

    def formula(self):
        """Return chemical formula string e.g. H2O"""
        parts = []
        for sym, cnt in sorted(self.formula_dict.items()):
            parts.append(sym + (str(cnt) if cnt > 1 else ""))
        return "".join(parts)

    def n_atoms(self):
        return sum(self.formula_dict.values())

    def n_electrons(self):
        total = 0
        for sym, cnt in self.formula_dict.items():
            total += get_element(sym)["Z"] * cnt
        return total

    def is_polar(self):
        """Crude polarity flag — True if more than one distinct element."""
        return len(self.formula_dict) > 1

    def __repr__(self):
        return f"Molecule({self.formula()}, M={self.molar_mass:.3f} g/mol)"


# Common molecules library
WATER   = Molecule("Water",   {"H":2,"O":1},  [("H","O",1),("O","H",1)])
CO2     = Molecule("Carbon dioxide",{"C":1,"O":2},[("C","O",2),("C","O",2)])
NH3     = Molecule("Ammonia", {"N":1,"H":3},  [("N","H",1)]*3)
CH4     = Molecule("Methane", {"C":1,"H":4},  [("C","H",1)]*4)
O2      = Molecule("Oxygen",  {"O":2},        [("O","O",2)])
N2      = Molecule("Nitrogen",{"N":2},        [("N","N",3)])
GLUCOSE = Molecule("Glucose", {"C":6,"H":12,"O":6})
ETHANOL = Molecule("Ethanol", {"C":2,"H":6,"O":1})
NaCl    = Molecule("Salt",    {"Na":1,"Cl":1},[("Na","Cl",1)])

def molecules_library():
    """Return dict of pre-built common molecules."""
    return {"H2O":WATER,"CO2":CO2,"NH3":NH3,"CH4":CH4,
            "O2":O2,"N2":N2,"glucose":GLUCOSE,
            "ethanol":ETHANOL,"NaCl":NaCl}

def van_der_waals_pressure(n, T, V, a, b, R=8.314):
    """(P + a*n^2/V^2)*(V - n*b) = n*R*T  -> P"""
    import numpy as np
    n,T,V,a,b = map(float,[n,T,V,a,b])
    return (n*R*T/(V-n*b)) - a*n**2/V**2




import math
import numpy as np

# ─────────────────────────────────────────────────────────────────
#  PKTron HHL Algorithm & Quantum Linear System Solvers
#  Memory safety: all matrices max 8x8, circuits max 10 qubits.
# ─────────────────────────────────────────────────────────────────


# ═══════════════════════════════════════════════════════════════
#  Validation
# ═══════════════════════════════════════════════════════════════

def _validate(A, b):
    A = np.array(A, dtype=complex)
    b = np.array(b, dtype=complex)
    assert A.ndim == 2,                        "A must be 2-D"
    assert A.shape[0] == A.shape[1],           "A must be square"
    assert b.ndim == 1,                        "b must be 1-D"
    assert b.shape[0] == A.shape[0],           "b length must match A dimension"
    assert A.shape[0] <= 8,                    f"A is {A.shape[0]}x{A.shape[0]} — max 8x8 for 12 GB safety"
    assert np.allclose(A, A.conj().T, atol=1e-8), "A must be Hermitian"
    assert not np.allclose(b, 0),              "b must be non-zero"
    return A, b


# ═══════════════════════════════════════════════════════════════
#  Result containers
# ═══════════════════════════════════════════════════════════════

class HHLResult:
    def __init__(self, A, b, x_q, x_c, residual, rel_err,
                 eigenvalues, eigenvectors, kappa, C,
                 b_coeffs, circuit, n_clock, sp_angles, steps):
        self.A = A;               self.b = b
        self.x_quantum = x_q;    self.x_classical = x_c
        self.residual = residual; self.relative_error = rel_err
        self.eigenvalues = eigenvalues; self.eigenvectors = eigenvectors
        self.condition_number = kappa;  self.C = C
        self.b_coeffs = b_coeffs;      self.circuit = circuit
        self.n_clock = n_clock;         self.state_prep_angles = sp_angles
        self.steps = steps;             self.success = residual < 1e-6

    def summary(self):
        sep = "=" * 60
        lines = [sep, "  HHL — Quantum Linear System Solver", sep]
        n = self.A.shape[0]
        lines += [
            f"  System size      : {n} x {n}",
            f"  Condition number : {self.condition_number:.4f}",
            f"  Eigenvalues      : {np.round(self.eigenvalues, 6).tolist()}",
            f"  Normalisation C  : {self.C:.6f}",
            f"  Clock qubits     : {self.n_clock}",
            f"  Circuit qubits   : {self.circuit.num_qubits}",
            f"  Circuit gates    : {len(self.circuit.gate_history)}",
            "",
            f"  b (input)        : {np.round(self.b.real, 6).tolist()}",
            f"  x_classical      : {np.round(self.x_classical, 6).tolist()}",
            f"  x_quantum        : {np.round(self.x_quantum, 6).tolist()}",
            f"  Residual ||Ax-b||: {self.residual:.4e}",
            f"  Relative error   : {self.relative_error:.4e}",
            f"  Status           : {'PASSED' if self.success else 'HIGH RESIDUAL'}",
        ]
        lines.append(sep)
        return "\n".join(lines)

    def __repr__(self):
        return (f"<HHLResult {self.A.shape[0]}x{self.A.shape[0]} "
                f"residual={self.residual:.2e} success={self.success}>")


class QLSResult:
    def __init__(self, method, A, b, x_quantum, x_classical,
                 residual, rel_err, steps):
        self.method = method;           self.A = A; self.b = b
        self.x_quantum = x_quantum;    self.x_classical = x_classical
        self.residual = residual;       self.relative_error = rel_err
        self.steps = steps
        # VQLS is iterative/approximate — looser threshold
        tol = 0.05 if method == "VQLS" else 1e-6
        self.success = residual < tol

    def summary(self):
        sep = "=" * 60
        lines = [sep, f"  QLS Solver: {self.method}", sep]
        lines += [
            f"  b (input)      : {np.round(self.b.real, 6).tolist()}",
            f"  x_classical    : {np.round(self.x_classical, 6).tolist()}",
            f"  x_quantum      : {np.round(self.x_quantum, 6).tolist()}",
            f"  Residual       : {self.residual:.4e}",
            f"  Relative error : {self.relative_error:.4e}",
            f"  Status         : {'PASSED' if self.success else 'FAILED'}",
        ]
        lines.append(sep)
        return "\n".join(lines)

    def __repr__(self):
        return (f"<QLSResult method={self.method} "
                f"residual={self.residual:.2e} success={self.success}>")


# ═══════════════════════════════════════════════════════════════
#  State-prep angles
# ═══════════════════════════════════════════════════════════════

def _state_prep_angles(b_norm):
    n = len(b_norm)
    angles = []
    cumulative = np.cumsum(np.abs(b_norm) ** 2)
    for i in range(n - 1):
        denom = 1.0 - (cumulative[i - 1] if i > 0 else 0.0)
        if denom < 1e-12:
            angles.append(0.0)
        else:
            ratio = min(abs(b_norm[i]) ** 2 / denom, 1.0)
            angles.append(2 * math.acos(math.sqrt(max(1 - ratio, 0))))
    angles.append(0.0)
    return np.array(angles)


# ═══════════════════════════════════════════════════════════════
#  Circuit builder (symbolic, max 10 qubits)
# ═══════════════════════════════════════════════════════════════

def _build_hhl_circuit(A, b_state, eigenvalues, C, n_clock):
    from .quantum import QuantumCircuit
    n       = A.shape[0]
    n_state = max(1, math.ceil(math.log2(max(n, 2))))
    total   = min(n_clock + 1 + n_state, 10)
    ancilla = n_clock
    qc      = QuantumCircuit(total)

    # Stage 1: state prep
    qc.gate_history.append(f"StatePrep |b>  Ry angles on state={n_clock+1}..{total-1}")
    for i in range(n_clock + 1, total):
        idx   = min(i - n_clock - 1, len(b_state) - 1)
        angle = float(2 * math.asin(min(abs(b_state[idx]), 1.0)))
        qc.rz(angle, i)

    # Stage 2: QPE — H on clock
    for q in range(n_clock):
        qc.h(q)

    # Stage 2: controlled e^{iAt}
    for k in range(n_clock):
        qc.gate_history.append(
            f"Controlled-e^{{iA*{2*math.pi*(2**k):.4f}}} "
            f"ctrl={k} state={n_clock+1}..{total-1}"
        )

    # Stage 2: inverse QFT
    for i in range(n_clock // 2):
        qc.swap(i, n_clock - 1 - i)
    for j in range(n_clock):
        qc.h(j)
        for kk in range(j + 1, n_clock):
            angle = -math.pi / (2 ** (kk - j))
            qc.rz(angle, j)
            qc.gate_history.append(f"CRZ({angle:.4f}) ctrl={kk} target={j}  [IQFT]")

    # Stage 3: eigenvalue inversion
    for lam in eigenvalues:
        if abs(lam) > 1e-10:
            theta = 2 * math.asin(min(abs(C / lam), 1.0))
            qc.gate_history.append(
                f"Controlled-Ry({theta:.4f}) [1/lambda={1/lam:.4f}] ancilla={ancilla}"
            )
    qc.rz(0.0, ancilla)

    # Stage 4: uncompute QPE
    qc.gate_history.append("QFT on clock  [uncompute QPE]")
    for j in range(n_clock - 1, -1, -1):
        for kk in range(n_clock - 1, j, -1):
            angle = math.pi / (2 ** (kk - j))
            qc.rz(angle, j)
        qc.h(j)
    for i in range(n_clock // 2):
        qc.swap(i, n_clock - 1 - i)

    qc.gate_history.append(f"Measure ancilla={ancilla} post-select |1>")
    qc.gate_history.append("Measure state register -> x")
    return qc


# ═══════════════════════════════════════════════════════════════
#  Core: hhl_solve
# ═══════════════════════════════════════════════════════════════

def hhl_solve(A, b, n_clock=None, verbose=True):
    """
    Solve A x = b using the HHL algorithm.

    Parameters
    ----------
    A       : Hermitian array (N x N), max 8x8
    b       : array (N,)
    n_clock : QPE clock qubits (auto-sized if None, capped at 6)
    verbose : bool

    Returns
    -------
    HHLResult — .x_quantum, .x_classical, .residual, .relative_error,
                .eigenvalues, .condition_number, .circuit, .n_clock,
                .b_coeffs, .C, .state_prep_angles
    """
    steps = []
    def log(msg):
        steps.append(msg)
        if verbose: print("  " + msg)

    A, b = _validate(A, b)
    n    = A.shape[0]
    log(f"System: {n}x{n}  Hermitian")

    b_norm_val = float(np.linalg.norm(b))
    b_state    = b / b_norm_val
    sp_angles  = _state_prep_angles(b_state.real)
    log(f"||b|| = {b_norm_val:.6f}")
    log(f"State-prep Ry angles: {np.round(sp_angles, 4).tolist()}")

    eigenvalues, eigenvectors = np.linalg.eigh(A)
    log(f"Eigenvalues of A: {np.round(eigenvalues, 6).tolist()}")

    abs_ev  = np.abs(eigenvalues)
    nonzero = abs_ev[abs_ev > 1e-10]
    kappa   = float(nonzero.max() / nonzero.min()) if len(nonzero) > 1 else 1.0
    log(f"Condition number kappa = {kappa:.4f}")

    if n_clock is None:
        n_clock = min(max(3, math.ceil(math.log2(kappa + 1)) + 1), 6)
    else:
        n_clock = min(int(n_clock), 6)
    log(f"QPE clock qubits: {n_clock}  (auto-sized from kappa, capped at 6)")

    b_coeffs = eigenvectors.conj().T @ b_state
    log(f"|b> in eigenbasis: {np.round(np.abs(b_coeffs), 4).tolist()}")

    C = float(nonzero.min())
    log(f"Normalisation constant C = {C:.6f}")

    inv_coeffs = np.where(
        np.abs(eigenvalues) > 1e-10,
        b_coeffs * C / eigenvalues,
        0.0 + 0.0j
    )
    log("Eigenvalue inversion complete  (ancilla post-selection simulated)")

    x_raw     = eigenvectors @ inv_coeffs
    x_quantum = (x_raw * b_norm_val / C).real
    log(f"x_quantum (raw): {np.round(x_quantum, 6).tolist()}")

    circuit = _build_hhl_circuit(A, b_state, eigenvalues, C, n_clock)
    log(f"HHL circuit: {circuit.num_qubits} qubits, {len(circuit.gate_history)} gates")

    x_classical = np.linalg.solve(A, b).real
    residual    = float(np.linalg.norm(A @ x_quantum - b.real))
    rel_err     = float(np.linalg.norm(x_quantum - x_classical) /
                        (np.linalg.norm(x_classical) + 1e-14))
    log(f"Residual  ||Ax-b||          = {residual:.4e}")
    log(f"Relative error ||xq-xc||/||xc|| = {rel_err:.4e}")

    return HHLResult(A, b, x_quantum, x_classical, residual, rel_err,
                     eigenvalues, eigenvectors, kappa, C,
                     b_coeffs, circuit, n_clock, sp_angles, steps)


# ═══════════════════════════════════════════════════════════════
#  VQLS
# ═══════════════════════════════════════════════════════════════

def variational_qls(A, b, iterations=120, verbose=True):
    """
    Variational Quantum Linear Solver (VQLS).
    Minimises ||A|x> - |b>|| via parameter-shift gradient descent.
    Returns QLSResult (.success threshold = 0.05 residual).
    """
    steps = []
    def log(msg):
        steps.append(msg)
        if verbose: print("  " + msg)

    A, b  = _validate(A, b)
    A_r   = A.real.astype(float)
    b_r   = b.real.astype(float)
    n     = A_r.shape[0]
    log(f"VQLS: {n}x{n} system, {iterations} iterations")

    b_state = b_r / (np.linalg.norm(b_r) + 1e-14)
    rng     = np.random.default_rng(42)
    theta   = rng.uniform(0, 2 * math.pi, n)
    lr      = 0.1
    best_cost, best_theta = float("inf"), theta.copy()

    for it in range(iterations):
        x_t = np.cos(theta)
        nm  = np.linalg.norm(x_t)
        x_t = x_t / nm if nm > 1e-12 else np.ones(n) / math.sqrt(n)

        Ax   = A_r @ x_t
        num  = float(b_state @ Ax) ** 2
        den  = float(np.linalg.norm(Ax)) ** 2 * float(np.linalg.norm(b_state)) ** 2
        cost = 1.0 - num / (den + 1e-12)

        if cost < best_cost:
            best_cost, best_theta = cost, theta.copy()

        grad = np.zeros(n)
        for j in range(n):
            for sign, s in [(+1, +math.pi/2), (-1, -math.pi/2)]:
                tp = theta.copy(); tp[j] += s
                xp = np.cos(tp); xp /= (np.linalg.norm(xp) + 1e-12)
                Axp = A_r @ xp
                cp  = 1.0 - (b_state @ Axp) ** 2 / (
                      np.linalg.norm(Axp) ** 2 * np.linalg.norm(b_state) ** 2 + 1e-12)
                grad[j] += sign * cp / 2
        theta -= lr * grad

        if (it + 1) % 30 == 0:
            log(f"  iter {it+1:3d}/{iterations}  cost={cost:.6f}")

    x_q  = np.cos(best_theta)
    x_c  = np.linalg.solve(A_r, b_r)
    scale = np.dot(x_q, x_c) / (np.dot(x_q, x_q) + 1e-14)
    x_q  *= scale

    residual = float(np.linalg.norm(A_r @ x_q - b_r))
    rel_err  = float(np.linalg.norm(x_q - x_c) / (np.linalg.norm(x_c) + 1e-14))
    log(f"Final cost={best_cost:.6f}  residual={residual:.4e}  rel_err={rel_err:.4e}")
    return QLSResult("VQLS", A, b, x_q, x_c, residual, rel_err, steps)


# ═══════════════════════════════════════════════════════════════
#  Quantum Conjugate Gradient
# ═══════════════════════════════════════════════════════════════

def quantum_conjugate_gradient(A, b, tol=1e-10, max_iter=200, verbose=True):
    """
    Quantum-inspired Conjugate Gradient. Exact for SPD A.
    Converges in at most N iterations.
    """
    steps = []
    def log(msg):
        steps.append(msg)
        if verbose: print("  " + msg)

    A, b = _validate(A, b)
    A_r  = A.real.astype(float)
    b_r  = b.real.astype(float)
    n    = A_r.shape[0]
    log(f"QCG: {n}x{n} system  tol={tol}")

    x      = np.zeros(n)
    r      = b_r - A_r @ x
    p      = r.copy()
    rs_old = float(r @ r)
    log(f"Initial residual: {math.sqrt(rs_old):.4e}")

    for i in range(min(max_iter, n * 10)):
        Ap     = A_r @ p
        denom  = float(p @ Ap)
        if abs(denom) < 1e-14: break
        alpha  = rs_old / denom
        x      = x + alpha * p
        r      = r - alpha * Ap
        rs_new = float(r @ r)
        res    = math.sqrt(rs_new)

        if (i + 1) % 20 == 0 or res < tol:
            log(f"  iter {i+1}  residual={res:.4e}")
        if res < tol:
            log(f"  Converged at iteration {i+1}")
            break

        p      = r + (rs_new / (rs_old + 1e-14)) * p
        rs_old = rs_new

    x_c      = np.linalg.solve(A_r, b_r)
    residual = float(np.linalg.norm(A_r @ x - b_r))
    rel_err  = float(np.linalg.norm(x - x_c) / (np.linalg.norm(x_c) + 1e-14))
    log(f"Final residual={residual:.4e}  rel_err={rel_err:.4e}")
    return QLSResult("QuantumCG", A, b, x, x_c, residual, rel_err, steps)


# ═══════════════════════════════════════════════════════════════
#  Randomised Kaczmarz
# ═══════════════════════════════════════════════════════════════

def randomised_kaczmarz(A, b, iterations=500, verbose=True):
    """
    Randomised Kaczmarz row-action solver.
    Converges for any consistent Hermitian system.
    """
    steps = []
    def log(msg):
        steps.append(msg)
        if verbose: print("  " + msg)

    A, b = _validate(A, b)
    A_r  = A.real.astype(float)
    b_r  = b.real.astype(float)
    n    = A_r.shape[0]
    log(f"Kaczmarz: {n}x{n} system, {iterations} iterations")

    row_norms_sq = np.sum(A_r ** 2, axis=1)
    probs        = row_norms_sq / row_norms_sq.sum()
    x            = np.zeros(n)
    rng          = np.random.default_rng(0)

    for it in range(iterations):
        i     = int(rng.choice(n, p=probs))
        ai    = A_r[i]
        denom = row_norms_sq[i]
        if denom < 1e-14: continue
        x     = x + ((b_r[i] - ai @ x) / denom) * ai

        if (it + 1) % 100 == 0:
            res = float(np.linalg.norm(A_r @ x - b_r))
            log(f"  iter {it+1}  residual={res:.4e}")

    x_c      = np.linalg.solve(A_r, b_r)
    residual = float(np.linalg.norm(A_r @ x - b_r))
    rel_err  = float(np.linalg.norm(x - x_c) / (np.linalg.norm(x_c) + 1e-14))
    log(f"Final residual={residual:.4e}  rel_err={rel_err:.4e}")
    return QLSResult("Kaczmarz", A, b, x, x_c, residual, rel_err, steps)


# ═══════════════════════════════════════════════════════════════
#  Quantum-Preconditioned GMRES (FIXED)
# ═══════════════════════════════════════════════════════════════

def quantum_preconditioned_gmres(A, b, max_iter=50, tol=1e-10, verbose=True):
    """
    Quantum-Preconditioned GMRES.
    Uses diagonal preconditioner P = diag(1/sqrt(|A_ii|)),
    then solves the preconditioned system with full GMRES.
    Works for any invertible Hermitian A.
    """
    steps = []
    def log(msg):
        steps.append(msg)
        if verbose: print("  " + msg)

    A, b = _validate(A, b)
    A_r  = A.real.astype(float)
    b_r  = b.real.astype(float)
    n    = A_r.shape[0]
    log(f"Q-GMRES: {n}x{n} system  max_iter={max_iter}")

    # Diagonal preconditioner
    diag       = np.abs(np.diag(A_r))
    diag       = np.where(diag > 1e-10, diag, 1.0)
    P_inv      = np.diag(1.0 / np.sqrt(diag))
    log(f"Preconditioner: diag({np.round(np.diag(P_inv), 4).tolist()})")

    # Preconditioned system: (P A P) y = P b,  x = P y
    PAP = P_inv @ A_r @ P_inv
    Pb  = P_inv @ b_r

    # ── Full GMRES on PAP y = Pb ─────────────────────────────
    m   = min(max_iter, n)
    y   = np.zeros(n)
    r0  = Pb - PAP @ y
    b0  = float(np.linalg.norm(r0))

    if b0 < tol:
        x = P_inv @ y
        x_c      = np.linalg.solve(A_r, b_r)
        residual = float(np.linalg.norm(A_r @ x - b_r))
        rel_err  = float(np.linalg.norm(x - x_c) / (np.linalg.norm(x_c) + 1e-14))
        log("Converged immediately")
        return QLSResult("Q-GMRES", A, b, x, x_c, residual, rel_err, steps)

    # Arnoldi basis Q (n x m+1), upper Hessenberg H (m+1 x m)
    Q  = np.zeros((n, m + 1))
    H  = np.zeros((m + 1, m))
    Q[:, 0] = r0 / b0

    # Givens rotation accumulators
    cs = np.zeros(m)
    sn = np.zeros(m)
    g  = np.zeros(m + 1)
    g[0] = b0

    solved = False
    j_end  = m

    for j in range(m):
        # Arnoldi step
        w = PAP @ Q[:, j]
        for i in range(j + 1):
            H[i, j] = float(Q[:, i] @ w)
            w        = w - H[i, j] * Q[:, i]
        H[j + 1, j] = float(np.linalg.norm(w))

        if H[j + 1, j] > 1e-12:
            Q[:, j + 1] = w / H[j + 1, j]

        # Apply previous Givens rotations to new column
        for i in range(j):
            tmp         =  cs[i] * H[i, j] + sn[i] * H[i + 1, j]
            H[i + 1, j] = -sn[i] * H[i, j] + cs[i] * H[i + 1, j]
            H[i, j]     = tmp

        # Compute new Givens rotation
        denom = math.sqrt(H[j, j] ** 2 + H[j + 1, j] ** 2)
        if denom < 1e-14:
            cs[j], sn[j] = 1.0, 0.0
        else:
            cs[j] = H[j, j]     / denom
            sn[j] = H[j + 1, j] / denom

        # Apply to H and g
        H[j, j]     =  cs[j] * H[j, j] + sn[j] * H[j + 1, j]
        H[j + 1, j] = 0.0
        g[j + 1]    = -sn[j] * g[j]
        g[j]        =  cs[j] * g[j]

        res_j = abs(g[j + 1])
        if (j + 1) % 10 == 0:
            log(f"  iter {j+1}  residual={res_j:.4e}")

        if res_j < tol:
            log(f"  Converged at iteration {j+1}")
            j_end  = j + 1
            solved = True
            break

    if not solved:
        j_end = m

    # Back-substitution on upper triangular H[:j_end, :j_end]
    Hj = H[:j_end, :j_end]
    gj = g[:j_end]
    try:
        y_sol = np.linalg.solve(Hj, gj)
    except np.linalg.LinAlgError:
        y_sol, _, _, _ = np.linalg.lstsq(Hj, gj, rcond=None)

    y = Q[:, :j_end] @ y_sol

    # Recover x = P y  (undo left preconditioning)
    x        = P_inv @ y
    x_c      = np.linalg.solve(A_r, b_r)
    residual = float(np.linalg.norm(A_r @ x - b_r))
    rel_err  = float(np.linalg.norm(x - x_c) / (np.linalg.norm(x_c) + 1e-14))
    log(f"Final residual={residual:.4e}  rel_err={rel_err:.4e}")
    return QLSResult("Q-GMRES", A, b, x, x_c, residual, rel_err, steps)




import numpy as np

# ─────────────────────────────────────────────────────────────────
#  PKTron Hybrid Quantum-Classical ML Module
#
#  Makes QuantumCircuit fully differentiable so it can be used
#  as a layer in PyTorch or TensorFlow neural networks.
#
#  Uses the parameter-shift rule for gradients:
#    dE/dθ = [E(θ+π/2) - E(θ-π/2)] / 2
#
#  Classes:
#    QuantumLayer      → PyTorch nn.Module wrapping a circuit
#    TFQuantumLayer    → TensorFlow/Keras layer wrapping a circuit
#    GradientEstimator → Compute gradients via parameter-shift
#    QNNClassifier     → Ready-to-use quantum neural network classifier
# ─────────────────────────────────────────────────────────────────


def _eval_circuit(circuit, param_dict, observable="Z", qubit=0):
    """Evaluate expectation value of a circuit with given parameters."""
    from .primitives import EstimatorV2
    bound = circuit.bind_parameters(param_dict)
    est = EstimatorV2(shots=512)
    result = est.run(bound, observable, qubit)
    return result.expectation_value


# ── Parameter-Shift Gradient ─────────────────────────────────────

class GradientEstimator:
    """
    Compute gradients of a parameterized circuit using
    the parameter-shift rule.

    dE/dθ = [E(θ + π/2) - E(θ - π/2)] / 2

    Usage:
        from pktron import Parameter, QuantumCircuit, GradientEstimator
        theta = Parameter("theta")
        qc = QuantumCircuit(1)
        qc.ry_param(0, theta)

        grad = GradientEstimator()
        gradients = grad.compute(qc, {theta: 1.0}, observable="Z", qubit=0)
        print(gradients)  # {theta: gradient_value}
    """

    def __init__(self, shift=np.pi/2):
        self.shift = shift

    def compute(self, circuit, param_dict, observable="Z", qubit=0):
        """
        Compute gradient with respect to all parameters.
        Returns dict {param: gradient_value}.
        """
        from .parameters import Parameter
        gradients = {}

        for param, value in param_dict.items():
            # Forward shift
            params_plus = {p: (v + self.shift if p == param else v)
                          for p, v in param_dict.items()}
            e_plus = _eval_circuit(circuit, params_plus, observable, qubit)

            # Backward shift
            params_minus = {p: (v - self.shift if p == param else v)
                           for p, v in param_dict.items()}
            e_minus = _eval_circuit(circuit, params_minus, observable, qubit)

            gradients[param] = (e_plus - e_minus) / 2.0

        return gradients

    def compute_all_qubits(self, circuit, param_dict, num_qubits, observable="Z"):
        """Compute gradients for all qubits."""
        all_grads = {}
        for q in range(num_qubits):
            grads = self.compute(circuit, param_dict, observable, q)
            for p, g in grads.items():
                key = f"{p.name if hasattr(p,'name') else p}_q{q}"
                all_grads[key] = g
        return all_grads


# ── PyTorch Quantum Layer ────────────────────────────────────────

class QuantumLayer:
    """
    A quantum circuit usable as a PyTorch neural network layer.
    Uses parameter-shift rule for backpropagation.

    Usage (with PyTorch):
        import torch
        from pktron import QuantumCircuit, Parameter, QuantumLayer

        theta = Parameter("theta")
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.ry_param(0, theta)
        qc.cnot(0, 1)

        qlayer = QuantumLayer(qc, [theta], observable="Z", qubit=0)
        output = qlayer.forward([0.5])      # forward pass
        grads  = qlayer.backward([0.5])     # gradients
    """

    def __init__(self, circuit, parameters, observable="Z", qubit=0, shots=512):
        self.circuit = circuit
        self.parameters = parameters
        self.observable = observable
        self.qubit = qubit
        self.shots = shots
        self._grad_estimator = GradientEstimator()
        self._torch_available = self._check_torch()

    def _check_torch(self):
        try:
            import torch
            return True
        except ImportError:
            return False

    def forward(self, param_values):
        """Forward pass: compute expectation value."""
        param_dict = {p: float(v) for p, v in zip(self.parameters, param_values)}
        return _eval_circuit(self.circuit, param_dict, self.observable, self.qubit)

    def backward(self, param_values):
        """Backward pass: compute gradients via parameter-shift."""
        param_dict = {p: float(v) for p, v in zip(self.parameters, param_values)}
        grads = self._grad_estimator.compute(
            self.circuit, param_dict, self.observable, self.qubit
        )
        return [grads[p] for p in self.parameters]

    def as_torch_function(self):
        """
        Returns a PyTorch autograd Function class for this layer.
        Use this to plug directly into a torch.nn.Module.
        """
        if not self._torch_available:
            raise ImportError("PyTorch not installed. pip install torch")

        import torch
        layer = self  # capture self

        class _QFunc(torch.autograd.Function):
            @staticmethod
            def forward(ctx, params):
                values = params.detach().numpy().tolist()
                ctx.save_for_backward(params)
                result = layer.forward(values)
                return torch.tensor([result], dtype=torch.float64)

            @staticmethod
            def backward(ctx, grad_output):
                params, = ctx.saved_tensors
                values = params.detach().numpy().tolist()
                grads = layer.backward(values)
                grad_tensor = torch.tensor(grads, dtype=torch.float64)
                return grad_output * grad_tensor

        return _QFunc.apply

    def as_torch_module(self):
        """Returns a torch.nn.Module wrapping this quantum layer."""
        if not self._torch_available:
            raise ImportError("PyTorch not installed. pip install torch")
        import torch
        import torch.nn as nn
        qfunc = self.as_torch_function()
        n_params = len(self.parameters)

        class _QModule(nn.Module):
            def __init__(self):
                super().__init__()
                self.params = nn.Parameter(
                    torch.randn(n_params, dtype=torch.float64)
                )
            def forward(self, x=None):
                return qfunc(self.params)

        return _QModule()


# ── TensorFlow Quantum Layer ─────────────────────────────────────

class TFQuantumLayer:
    """
    A quantum circuit usable as a TensorFlow/Keras layer.
    Uses parameter-shift rule for gradients via tf.custom_gradient.

    Usage:
        from pktron import QuantumCircuit, Parameter, TFQuantumLayer
        theta = Parameter("theta")
        qc = QuantumCircuit(1)
        qc.ry_param(0, theta)

        tflayer = TFQuantumLayer(qc, [theta])
        result = tflayer([0.5])   # forward pass
    """

    def __init__(self, circuit, parameters, observable="Z", qubit=0):
        self.circuit = circuit
        self.parameters = parameters
        self.observable = observable
        self.qubit = qubit
        self._grad_estimator = GradientEstimator()
        self._tf_available = self._check_tf()

    def _check_tf(self):
        try:
            import tensorflow as tf
            return True
        except ImportError:
            return False

    def __call__(self, param_values):
        """Forward pass."""
        vals = [float(v) for v in param_values]
        return self.forward(vals)

    def forward(self, param_values):
        param_dict = {p: v for p, v in zip(self.parameters, param_values)}
        return _eval_circuit(self.circuit, param_dict, self.observable, self.qubit)

    def as_tf_function(self):
        """Returns a TF function with custom gradient support."""
        if not self._tf_available:
            raise ImportError("TensorFlow not installed. pip install tensorflow")
        import tensorflow as tf
        layer = self

        @tf.custom_gradient
        def _qfunc(params):
            values = [float(p) for p in params.numpy()]
            result = layer.forward(values)

            def grad(dy):
                param_dict = {p: v for p, v in zip(layer.parameters, values)}
                grads = layer._grad_estimator.compute(
                    layer.circuit, param_dict, layer.observable, layer.qubit
                )
                grad_vals = [grads[p] for p in layer.parameters]
                return dy * tf.constant(grad_vals, dtype=tf.float64)

            return tf.constant([result], dtype=tf.float64), grad

        return _qfunc


# ── QNN Classifier ───────────────────────────────────────────────

class QNNClassifier:
    """
    A ready-to-use quantum neural network binary classifier.
    Uses a variational ansatz + gradient descent.

    Usage:
        from pktron import QNNClassifier

        qnn = QNNClassifier(num_qubits=2, layers=3, shots=512)
        qnn.fit(X_train, y_train, epochs=20, lr=0.1)
        preds = qnn.predict(X_test)
        print(f"Accuracy: {qnn.score(X_test, y_test):.2%}")
    """

    def __init__(self, num_qubits=2, layers=3, shots=512, observable="Z", qubit=0):
        self.num_qubits = num_qubits
        self.layers = layers
        self.shots = shots
        self.observable = observable
        self.qubit = qubit
        self._grad = GradientEstimator()
        self._build_circuit()

    def _build_circuit(self):
        from .quantum import QuantumCircuit
        from .parameters import ParameterVector
        self._params = ParameterVector("w", 2 * self.num_qubits * self.layers)
        self._circuit = QuantumCircuit(self.num_qubits)
        idx = 0
        for _ in range(self.layers):
            for q in range(self.num_qubits):
                self._circuit.ry_param(q, self._params[idx])
                self._circuit.rz_param(q, self._params[idx + self.num_qubits])
                idx += 1
            for q in range(self.num_qubits - 1):
                self._circuit.cnot(q, q + 1)
        self._weights = np.random.uniform(-np.pi, np.pi, len(self._params.params))

    def _param_dict(self, weights):
        return {self._params[i]: float(weights[i])
                for i in range(len(self._params.params))}

    def _encode(self, x):
        """Encode input x into circuit rotation angles."""
        # Simple angle encoding: scale features to [-pi, pi]
        if hasattr(x, "__len__"):
            scale = np.pi * np.tanh(np.array(x, dtype=float))
        else:
            scale = [float(x) * np.pi]
        return scale

    def _forward(self, x):
        from .quantum import QuantumCircuit
        # Build data-encoded + variational circuit
        qc = QuantumCircuit(self.num_qubits)
        features = self._encode(x)
        for q in range(min(self.num_qubits, len(features))):
            qc.ry(q, features[q])
        # Apply variational part
        bound = self._circuit.bind_parameters(self._param_dict(self._weights))
        for gate in bound.gate_history:
            qc.gate_history.append(gate)
        qc._allocate_state()
        # Apply encoding state
        from .gates import ry
        for q in range(min(self.num_qubits, len(features))):
            ry(qc, q, features[q])
        from .primitives import EstimatorV2
        est = EstimatorV2(shots=self.shots)
        r = est.run(bound, self.observable, self.qubit)
        return r.expectation_value

    def fit(self, X, y, epochs=10, lr=0.1):
        """Train the QNN using gradient descent."""
        losses = []
        for epoch in range(epochs):
            total_loss = 0.0
            grad_accum = np.zeros_like(self._weights)

            for xi, yi in zip(X, y):
                pred = _eval_circuit(
                    self._circuit,
                    self._param_dict(self._weights),
                    self.observable,
                    self.qubit
                )
                label = 1.0 if yi == 1 else -1.0
                loss = (pred - label) ** 2
                total_loss += loss

                # Gradient via parameter shift
                for i, p in enumerate(self._params.params):
                    pd = self._param_dict(self._weights)
                    pd_plus = {pp: (v + np.pi/2 if pp == p else v)
                               for pp, v in pd.items()}
                    pd_minus = {pp: (v - np.pi/2 if pp == p else v)
                                for pp, v in pd.items()}
                    e_plus  = _eval_circuit(self._circuit, pd_plus,
                                            self.observable, self.qubit)
                    e_minus = _eval_circuit(self._circuit, pd_minus,
                                            self.observable, self.qubit)
                    grad_accum[i] += 2 * (pred - label) * (e_plus - e_minus) / 2

            self._weights -= lr * grad_accum / len(X)
            avg_loss = total_loss / len(X)
            losses.append(avg_loss)
            if (epoch + 1) % max(1, epochs // 5) == 0:
                print(f"   Epoch {epoch+1}/{epochs} — loss: {avg_loss:.4f}")

        return losses

    def predict_proba(self, X):
        """Return raw expectation values for each sample."""
        return [_eval_circuit(self._circuit, self._param_dict(self._weights),
                              self.observable, self.qubit) for _ in X]

    def predict(self, X):
        """Predict binary class labels."""
        probas = self.predict_proba(X)
        return [1 if p >= 0 else 0 for p in probas]

    def score(self, X, y):
        """Return accuracy."""
        preds = self.predict(X)
        return np.mean([p == yi for p, yi in zip(preds, y)])




import numpy as np
import re

try:
    import cupy as cp
    _t = cp.zeros(1)
    _t[0] = 1.0
    HAS_GPU = True
except Exception:
    cp = None
    HAS_GPU = False

def _xp():
    if not HAS_GPU:
        return np
    try:
        test = cp.zeros(1)
        test[0] = 1.0
        return cp
    except Exception:
        return np

def _to_np(arr):
    if HAS_GPU and cp is not None and isinstance(arr, cp.ndarray):
        return cp.asnumpy(arr)
    return np.array(arr)

# ── Gate matrices ────────────────────────────────────────────────
_SQ2 = np.sqrt(2.0)
GATE_LIB_NP = {
    "H":  np.array([[1,1],[1,-1]], dtype=complex) / _SQ2,
    "X":  np.array([[0,1],[1,0]], dtype=complex),
    "Y":  np.array([[0,-1j],[1j,0]], dtype=complex),
    "Z":  np.array([[1,0],[0,-1]], dtype=complex),
    "S":  np.array([[1,0],[0,1j]], dtype=complex),
    "T":  np.array([[1,0],[0,np.exp(1j*np.pi/4)]], dtype=complex),
}
TWO_QUBIT_NP = {
    "CNOT": np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]], dtype=complex),
    "CZ":   np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]], dtype=complex),
    "CY":   np.array([[1,0,0,0],[0,1,0,0],[0,0,0,-1j],[0,0,1j,0]], dtype=complex),
    "SWAP": np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]], dtype=complex),
}

# ── Parse gate_history into structured ops ────────────────────────
# Exact formats from gates.py:
#   "H on qubit 0"
#   "Rx(1.57) on qubit 0"
#   "CNOT control 0 target 1"
#   "CZ control 0 target 1"
#   "CY control 0 target 1"
#   "SWAP qubits 0 and 1"
#   "Toffoli control1 0 control2 1 target 2"

def _parse_history(circuit):
    ops = []
    for entry in circuit.gate_history:
        e = entry.strip()
        el = e.lower()

        # Two-qubit: CNOT / CZ / CY
        m = re.match(r"(cnot|cz|cy)\s+control\s+(\d+)\s+target\s+(\d+)", el)
        if m:
            ops.append((m.group(1).upper(), [int(m.group(2)), int(m.group(3))], []))
            continue

        # SWAP
        m = re.match(r"swap\s+qubits\s+(\d+)\s+and\s+(\d+)", el)
        if m:
            ops.append(("SWAP", [int(m.group(1)), int(m.group(2))], []))
            continue

        # Toffoli
        m = re.match(r"toffoli\s+control1\s+(\d+)\s+control2\s+(\d+)\s+target\s+(\d+)", el)
        if m:
            ops.append(("TOFFOLI", [int(m.group(1)), int(m.group(2)), int(m.group(3))], []))
            continue

        # Parameterized: Rx/Ry/Rz
        m = re.match(r"r([xyz])\(([-\d.]+)\)\s+on\s+qubit\s+(\d+)", el)
        if m:
            ops.append(("R"+m.group(1).upper(), [int(m.group(3))], [float(m.group(2))]))
            continue

        # Single-qubit no param: "H on qubit 0"
        m = re.match(r"([a-z†]+)\s+on\s+qubit\s+(\d+)", el)
        if m:
            gname = m.group(1).upper().replace("†","DG")
            ops.append((gname, [int(m.group(2))], []))
            continue

    return ops

# ── Apply single-qubit gate to statevector ────────────────────────
def _apply_1q(xp, sv, G, qubit, n):
    G_xp = xp.array(G, dtype=complex)
    sv = sv.reshape([2]*n)
    # contract: new[...P...] = G[P,p] * sv[...p...]
    sv = xp.tensordot(G_xp, sv, axes=[[1],[qubit]])
    # tensordot puts the new axis at position 0; move it back to qubit
    order = list(range(1, qubit+1)) + [0] + list(range(qubit+1, n))
    sv = xp.transpose(sv, order)
    return sv.reshape(2**n)

# ── Apply two-qubit gate to statevector ───────────────────────────
def _apply_2q(xp, sv, G4, q0, q1, n):
    G_xp = xp.array(G4.reshape(2,2,2,2), dtype=complex)
    sv = sv.reshape([2]*n)
    # contract over axes q0,q1 simultaneously
    sv = xp.tensordot(G_xp, sv, axes=[[2,3],[q0,q1]])
    # result shape: (2,2, ...n-2 axes...)
    # axes 0,1 correspond to output q0,q1; rest are in original order
    # build permutation to put them back
    remaining = [i for i in range(n) if i not in (q0,q1)]
    # current order: [out_q0=0, out_q1=1, remaining...]
    # target order: insert 0->q0, 1->q1 into remaining positions
    full = list(range(2, n))   # remaining axes (0-indexed after the 2 output axes)
    perm = [None] * n
    perm[q0] = 0
    perm[q1] = 1
    ri = 2
    for pos in remaining:
        perm[pos] = ri; ri += 1
    sv = xp.transpose(sv, perm)
    return sv.reshape(2**n)

# ── Build rotation gates ──────────────────────────────────────────
def _rot_gate(axis, theta):
    c, s = np.cos(theta/2), np.sin(theta/2)
    if axis == "X": return np.array([[c,-1j*s],[-1j*s,c]], dtype=complex)
    if axis == "Y": return np.array([[c,-s],[s,c]], dtype=complex)
    if axis == "Z": return np.array([[np.exp(-1j*theta/2),0],[0,np.exp(1j*theta/2)]], dtype=complex)

# ── Run one circuit, return probability vector (on device) ────────
def _run_sv_probs(xp, circuit):
    n = circuit.num_qubits
    sv = xp.zeros(2**n, dtype=complex)
    sv[0] = 1.0 + 0j

    ops = _parse_history(circuit)
    for gate, qubits, params in ops:
        if gate in GATE_LIB_NP:
            sv = _apply_1q(xp, sv, GATE_LIB_NP[gate], qubits[0], n)
        elif gate in ("RX","RY","RZ"):
            th = params[0] if params else 0.0
            G = _rot_gate(gate[1], th)
            sv = _apply_1q(xp, sv, G, qubits[0], n)
        elif gate in TWO_QUBIT_NP and len(qubits) >= 2:
            sv = _apply_2q(xp, sv, TWO_QUBIT_NP[gate], qubits[0], qubits[1], n)
        # Toffoli and unknown gates: skip (rare in basic circuits)

    probs = xp.abs(sv)**2
    probs = probs / probs.sum()
    return probs

# ── GPUBatchedSampler ─────────────────────────────────────────────
class GPUBatchedSampler:
    """
    Run many circuits in a batch — all on GPU (or NumPy fallback).

    Usage:
        sampler = GPUBatchedSampler(shots=1024)
        result  = sampler.run(qc)
        results = sampler.run_batch([qc1, qc2, qc3])
    """
    def __init__(self, shots=1024):
        self.shots = shots
        self.xp = _xp()
        self.using_gpu = HAS_GPU

    def run(self, circuit, shots=None):
        shots = shots or self.shots
        xp = self.xp
        n = circuit.num_qubits
        probs_dev = _run_sv_probs(xp, circuit)
        probs_np = _to_np(probs_dev)
        probs_np = np.clip(probs_np, 0, None)
        probs_np /= probs_np.sum()
        outcomes = np.random.choice(2**n, shots, p=probs_np)
        counts = {format(k, f"0{n}b"): int(v)
                  for k, v in zip(*np.unique(outcomes, return_counts=True))}
        return {"counts": counts, "using_gpu": self.using_gpu}

    def run_batch(self, circuits, shots=None):
        return [self.run(qc, shots) for qc in circuits]

    def properties(self):
        return {"name": "GPUBatchedSampler",
                "backend": "CuPy-GPU" if self.using_gpu else "NumPy-CPU",
                "shots": self.shots}


# ── GPUBatchedEstimator ───────────────────────────────────────────
_PAULIS = {
    "I": np.eye(2,dtype=complex),
    "X": np.array([[0,1],[1,0]],dtype=complex),
    "Y": np.array([[0,-1j],[1j,0]],dtype=complex),
    "Z": np.array([[1,0],[0,-1]],dtype=complex),
}

class GPUBatchedEstimator:
    """
    Compute expectation values of Pauli strings on GPU.

    Usage:
        est = GPUBatchedEstimator()
        r   = est.run(qc, "ZZ")
        rs  = est.run_batch([(qc1,"ZZ"),(qc2,"XI")])
    """
    def __init__(self):
        self.xp = _xp()
        self.using_gpu = HAS_GPU

    def _build_obs(self, pauli_str, n):
        mat = np.array([[1.0+0j]])
        ps = pauli_str.upper().ljust(n,"I")[:n]
        for c in ps:
            mat = np.kron(mat, _PAULIS.get(c, _PAULIS["I"]))
        return self.xp.array(mat, dtype=complex)

    def run(self, circuit, pauli_str):
        xp = self.xp
        n = circuit.num_qubits
        sv = xp.zeros(2**n, dtype=complex); sv[0] = 1.0+0j
        ops = _parse_history(circuit)
        for gate, qubits, params in ops:
            if gate in GATE_LIB_NP:
                sv = _apply_1q(xp, sv, GATE_LIB_NP[gate], qubits[0], n)
            elif gate in ("RX","RY","RZ"):
                sv = _apply_1q(xp, sv, _rot_gate(gate[1], params[0] if params else 0.0), qubits[0], n)
            elif gate in TWO_QUBIT_NP and len(qubits) >= 2:
                sv = _apply_2q(xp, sv, TWO_QUBIT_NP[gate], qubits[0], qubits[1], n)
        O = self._build_obs(pauli_str, n)
        exp_val = float(xp.real(xp.conj(sv) @ O @ sv))
        return {"expectation_value": exp_val, "pauli": pauli_str,
                "using_gpu": self.using_gpu}

    def run_batch(self, jobs):
        return [self.run(qc, obs) for qc, obs in jobs]


# ── Convenience one-liner ─────────────────────────────────────────
def gpu_execute(circuit, shots=1024, observable=None):
    """
    One-line GPU execution.
        counts = gpu_execute(qc, shots=1024)
        exp    = gpu_execute(qc, observable="ZZ")
    """
    if observable:
        return GPUBatchedEstimator().run(circuit, observable)
    return GPUBatchedSampler(shots=shots).run(circuit)




"""PKTron Chemistry — Quantum Chemistry helpers."""
import numpy as np
_R = np.float32

h_eV   = _R(4.1357e-15)   # eV·s
hbar   = _R(1.0546e-34)   # J·s
me     = _R(9.109e-31)    # kg
a0     = _R(5.292e-11)    # Bohr radius m
E_h    = _R(27.211)       # Hartree in eV

def hydrogen_energy_level(n):
    """E_n = -13.6 / n^2 eV"""
    return float(-_R(13.6)/_R(n)**2)

def photon_energy_eV(wavelength_nm):
    """E = hc/lambda  in eV"""
    hc_eV_nm = _R(1239.8)
    return float(hc_eV_nm/_R(wavelength_nm))

def transition_wavelength_nm(n1, n2):
    """Wavelength of H-atom transition n1->n2 (nm)."""
    E1 = hydrogen_energy_level(n1)
    E2 = hydrogen_energy_level(n2)
    dE = abs(E2-E1)
    return float(_R(1239.8)/dE)

def hydrogen_wavefunction_1s(r_bohr):
    """Psi_1s(r) = (1/sqrt(pi))*exp(-r) in atomic units."""
    r = np.asarray(r_bohr, dtype=_R)
    return np.exp(-r)/np.sqrt(_R(np.pi))

def radial_probability_1s(r_bohr):
    """P(r) = 4*pi*r^2 * |psi|^2"""
    r   = np.asarray(r_bohr, dtype=_R)
    psi = hydrogen_wavefunction_1s(r)
    return 4*_R(np.pi)*r**2*psi**2

def molecular_orbital_energy_huckel(alpha, beta, k, n):
    """Huckel MO energy: E_k = alpha + 2*beta*cos(k*pi/(n+1))"""
    return float(_R(alpha)+2*_R(beta)*np.cos(_R(k)*_R(np.pi)/(_R(n)+1)))

def lennard_jones(r, epsilon, sigma):
    """LJ potential V = 4*eps*[(sigma/r)^12 - (sigma/r)^6]"""
    r = np.asarray(r, dtype=_R)
    sr6 = (_R(sigma)/r)**6
    return 4*_R(epsilon)*(sr6**2 - sr6)




import numpy as np
from fractions import Fraction
from scipy.optimize import minimize
from .gates import h, rx, ry, rz, cnot, cz, s, toffoli, t, x, swap
from .quantum import QuantumCircuit
try:
    import pennylane as qml
    _HAS_PENNYLANE = True
except ImportError:
    qml = None
    _HAS_PENNYLANE = False

def quantum_flip(circuit):
    circuit.h(0)
    if circuit.num_qubits > 1:
        circuit.cnot(0, 1)
    circuit.rx(0, np.pi/4)
    circuit.gate_history.append("Quantum Flip")

def super_quantum_dance(circuit):
    circuit.h(0)
    if circuit.num_qubits > 1:
        circuit.ry(1, np.pi/2)
        circuit.cnot(0, 1)
    if circuit.num_qubits > 2:
        circuit.toffoli(0, 1, 2)
    circuit.t(0)
    circuit.gate_history.append("Super Quantum Dance")

def quantum_treasure_hunt(circuit):
    target_state = 2**circuit.num_qubits - 1
    circuit.h(0)
    if circuit.num_qubits > 1:
        circuit.cnot(0, 1)
        circuit.ry(1, np.random.uniform(0, np.pi))
    circuit.rx(0, np.random.uniform(0, np.pi))
    outcome, probs = circuit.measure()
    score = probs[target_state] * 100
    circuit.gate_history.append(f"Quantum Treasure Hunt (Score: {score:.1f})")
    return score

def quantum_magic_spin(circuit):
    target_state = 0
    circuit.rx(0, np.random.uniform(0, 2*np.pi))
    if circuit.num_qubits > 1:
        circuit.ry(1, np.random.uniform(0, np.pi))
        circuit.cz(0, 1)
    circuit.s(0)
    outcome, probs = circuit.measure()
    score = probs[target_state] * 100
    circuit.gate_history.append(f"Quantum Magic Spin (Score: {score:.1f})")
    return score

def bb84_qkd(num_qubits, eve_intercept_prob=0.5, seed=None):
    if seed is not None:
        np.random.seed(seed)
    qc = QuantumCircuit(num_qubits, "statevector")
    alice_bits = np.random.randint(0, 2, num_qubits)
    alice_bases = np.random.randint(0, 2, num_qubits)
    eve_intercepts = np.random.choice([0,1], size=num_qubits, p=[1-eve_intercept_prob, eve_intercept_prob])
    eve_bases = np.random.randint(0, 2, num_qubits)
    bob_bases = np.random.randint(0, 2, num_qubits)
    for i in range(num_qubits):
        if alice_bits[i] == 1: qc.x(i)
        if alice_bases[i] == 1: qc.h(i)
    for i in range(num_qubits):
        if eve_intercepts[i]:
            if eve_bases[i] == 1: qc.h(i)
            if np.random.randint(0, 2) == 1: qc.x(i)
            if eve_bases[i] == 1: qc.h(i)
    for i in range(num_qubits):
        if bob_bases[i] == 1: qc.h(i)
    outcome, probs = qc.measure()
    measured_bits = [int(b) for b in format(outcome, f"0{qc.num_qubits}b")]
    alice_key, bob_key = [], []
    for i in range(num_qubits):
        if alice_bases[i] == bob_bases[i]:
            alice_key.append(alice_bits[i])
            bob_key.append(measured_bits[i])
    error_count = sum(a != b for a, b in zip(alice_key, bob_key))
    error_rate = error_count / len(alice_key) if alice_key else 0.0
    eve_detected = error_rate > 0.25
    qc.gate_history.append(f"BB84 QKD (Error rate: {error_rate:.2f}, Eve detected: {eve_detected})")
    return {"alice_key": alice_key, "bob_key": bob_key, "error_rate": error_rate,
            "eve_detected": eve_detected, "circuit": qc}

def grover_search(num_qubits, target=None, oracle_func=None, iterations=None, noise_prob=0.0):
    if num_qubits < 2:
        raise ValueError("Grover requires at least 2 qubits")
    qc = QuantumCircuit(num_qubits, "statevector")
    if iterations is None:
        iterations = int(np.pi / 4 * np.sqrt(2**num_qubits))
    def default_oracle():
        target_bin = format(target, f"0{num_qubits}b")
        for i, bit in enumerate(target_bin[::-1]):
            if bit == "0": qc.x(i)
        qc.h(num_qubits-1)
        if num_qubits > 2:
            qc.toffoli(0, 1, num_qubits-1)
        else:
            qc.cnot(0, 1)
        qc.h(num_qubits-1)
        for i, bit in enumerate(target_bin[::-1]):
            if bit == "0": qc.x(i)
    def diffusion():
        for i in range(num_qubits): qc.h(i); qc.x(i)
        qc.h(num_qubits-1)
        if num_qubits > 2:
            qc.toffoli(0, 1, num_qubits-1)
        else:
            qc.cnot(0, 1)
        qc.h(num_qubits-1)
        for i in range(num_qubits): qc.x(i); qc.h(i)
    for i in range(num_qubits): qc.h(i)
    for _ in range(iterations):
        if oracle_func is not None: oracle_func(qc)
        else: default_oracle()
        if np.random.random() < noise_prob: qc.x(np.random.randint(0, num_qubits))
        diffusion()
    outcome, probs = qc.measure()
    qc.gate_history.append(f"Grover Search (Target: {target})")
    return {"circuit": qc, "outcome": outcome, "probs": probs}

def qft(circuit, start=0, end=None):
    if end is None: end = circuit.num_qubits
    for i in range(start, end):
        for j in range(i+1, end):
            lam = np.pi / (2**(j-i))
            circuit.rz(i, lam/2)
            circuit.cnot(i, j)
            circuit.rz(j, -lam/2)
            circuit.cnot(i, j)
            circuit.rz(j, lam/2)
        circuit.h(i)
    for i in range(start, (start+end)//2):
        circuit.swap(i, end-1-(i-start))
    circuit.gate_history.append("Quantum Fourier Transform")

def phase_estimation(circuit, unitary, estimation_qubits, target_qubits):
    for q in estimation_qubits: circuit.h(q)
    for i, eq in enumerate(estimation_qubits):
        for _ in range(2**i): unitary(circuit, control=eq, targets=target_qubits)
    n_est = len(estimation_qubits)
    for i in range(n_est-1, -1, -1):
        for j in range(i):
            lam = -np.pi / (2**(i-j))
            circuit.rz(estimation_qubits[i], lam/2)
            circuit.cnot(estimation_qubits[i], estimation_qubits[j])
            circuit.rz(estimation_qubits[j], -lam/2)
            circuit.cnot(estimation_qubits[i], estimation_qubits[j])
            circuit.rz(estimation_qubits[j], lam/2)
        circuit.h(estimation_qubits[i])
    circuit.gate_history.append("Phase Estimation")
    outcome, probs = circuit.measure()
    return {"circuit": circuit, "phase": outcome/(2**n_est), "probs": probs}

def shor_period_finding(N, a, num_qubits=None):
    if num_qubits is None:
        num_qubits = 2 * int(np.ceil(np.log2(N))) + 1
    if num_qubits % 2 != 0: num_qubits += 1
    qc = QuantumCircuit(num_qubits, "statevector")
    counting_qubits = num_qubits // 2
    register_qubits = num_qubits - counting_qubits
    for i in range(counting_qubits): qc.h(i)
    qc.x(counting_qubits)
    period, factors = None, []
    for _ in range(1000):
        outcome, probs = qc.measure()
        measured = outcome >> register_qubits
        try:
            phase = measured / (2**counting_qubits)
            frac = Fraction(phase).limit_denominator(N)
            r = frac.denominator
            if r != 0 and r < N and pow(a, r, N) == 1:
                cand1 = np.gcd(pow(a, r//2, N)-1, N)
                cand2 = np.gcd(pow(a, r//2, N)+1, N)
                new_factors = [f for f in [cand1, cand2] if 1 < f < N]
                if new_factors:
                    period = r
                    factors = sorted(new_factors)
                    break
        except: pass
    qc.gate_history.append(f"Shor Period Finding (N={N}, a={a}, period={period})")
    return {"circuit": qc, "period": period, "factors": factors}

def vqe(num_qubits, hamiltonian=None, ansatz_depth=3, noise_prob=0.0, max_iter=300):
    if hamiltonian is None:
        hamiltonian = np.diag([1.0 if i==0 else -1.0 for i in range(2**num_qubits)])
    def ansatz(params, circuit):
        param_idx = 0
        for d in range(ansatz_depth):
            for i in range(num_qubits):
                circuit.ry(i, params[param_idx])
                circuit.rz(i, params[param_idx+num_qubits])
                param_idx += 1
            if num_qubits > 1:
                for i in range(num_qubits-1): circuit.cnot(i, i+1)
    def objective(params):
        qc = QuantumCircuit(num_qubits, "statevector")
        ansatz(params, qc)
        state = qc.get_state()
        return float(np.real(state.conj() @ hamiltonian @ state))
    init = np.random.uniform(0, np.pi, 2*num_qubits*ansatz_depth)
    result = minimize(objective, init, method="Powell", options={"maxiter": max_iter})
    qc = QuantumCircuit(num_qubits, "statevector")
    ansatz(result.x, qc)
    energy = objective(result.x)
    qc.gate_history.append(f"VQE (Energy: {energy:.4f})")
    return {"circuit": qc, "energy": energy, "params": result.x}

def qaoa(num_qubits, graph=None, p=1, noise_prob=0.0, max_iter=100):
    if graph is None:
        graph = [(0,1),(1,2)] if num_qubits >= 3 else [(0,1)]
    def cost_h(circuit, gamma):
        for edge in graph: circuit.cz(edge[0], edge[1])
    def mixer_h(circuit, beta):
        for i in range(num_qubits): circuit.rx(i, 2*beta)
    def objective(params):
        qc = QuantumCircuit(num_qubits, "statevector")
        for i in range(num_qubits): qc.h(i)
        for layer in range(p):
            cost_h(qc, params[2*layer])
            mixer_h(qc, params[2*layer+1])
        outcome, probs = qc.measure()
        cut = sum(probs[i]*sum(1 for u,v in graph if (i>>u)&1 != (i>>v)&1) for i in range(2**num_qubits))
        return -cut
    init = np.random.uniform(0, np.pi, 2*p)
    result = minimize(objective, init, method="COBYLA", options={"maxiter": max_iter})
    qc = QuantumCircuit(num_qubits, "statevector")
    for i in range(num_qubits): qc.h(i)
    for layer in range(p):
        cost_h(qc, result.x[2*layer])
        mixer_h(qc, result.x[2*layer+1])
    outcome, probs = qc.measure()
    cut = sum(probs[i]*sum(1 for u,v in graph if (i>>u)&1 != (i>>v)&1) for i in range(2**num_qubits))
    qc.gate_history.append(f"QAOA (Cut Value: {cut:.2f})")
    return {"circuit": qc, "cut_value": cut, "params": result.x}

def hhl(num_qubits, matrix=None, vector=None):
    if matrix is None: matrix = np.array([[1,-0.5],[-0.5,1]])
    if vector is None: vector = np.array([1,0])
    if num_qubits < 3: raise ValueError("HHL needs at least 3 qubits")
    if not np.allclose(matrix, matrix.conj().T): raise ValueError("Matrix must be Hermitian")
    n_state = int(np.log2(len(vector)))
    n_eigen = num_qubits - n_state - 1
    qc = QuantumCircuit(num_qubits, "statevector")
    ancilla = 0
    eigen_qubits = list(range(1, n_eigen+1))
    state_qubits = list(range(n_eigen+1, num_qubits))
    norm = np.linalg.norm(vector)
    if norm == 0: raise ValueError("Input vector cannot be zero")
    vector = vector / norm
    for i in range(n_state):
        if abs(vector[i]) > 1e-10:
            theta = 2*np.arccos(min(1, abs(vector[i])))
            qc.ry(state_qubits[i], theta)
    for i, eq in enumerate(eigen_qubits):
        for _ in range(2**i):
            theta = np.arccos(matrix[0,0] / np.linalg.norm(matrix))
            cos, sin = np.cos(theta/2), np.sin(theta/2)
            ry_gate = np.array([[cos,-sin],[sin,cos]])
            for st in state_qubits:
                dim = 4
                fg = np.eye(dim, dtype=complex)
                fg[2:,2:] = ry_gate
                qc._apply_gate(fg, [eq, st])
                qc.gate_history.append(f"Controlled-RY control {eq} target {st}")
    for i, q in enumerate(eigen_qubits):
        angle = np.pi / (2**(i+1))
        qc.cnot(q, ancilla)
        qc.ry(ancilla, angle)
        qc.cnot(q, ancilla)
    qc.normalize()
    outcome, probs = qc.measure()
    probs = probs / probs.sum()
    solution = np.array([probs[int("1"+"0"*n_eigen+format(i, f"0{n_state}b"), 2)]
                         if int("1"+"0"*n_eigen+format(i, f"0{n_state}b"), 2) < len(probs) else 0
                         for i in range(2**n_state)])
    sn = np.linalg.norm(solution)
    if sn > 1e-10: solution = solution / sn
    qc.gate_history.append(f"HHL (Solution norm: {np.linalg.norm(solution):.4f})")
    return {"circuit": qc, "solution": solution, "probs": probs}




import numpy as np

def _qft_matrix(n):
    N   = 2 ** n
    w   = np.exp(2j * np.pi / N)
    idx = np.arange(N)
    return np.power(w, np.outer(idx, idx)) / np.sqrt(N)

def qpe(unitary: np.ndarray, eigenstate: np.ndarray, n_ancilla: int) -> dict:
    """
    Standard Quantum Phase Estimation.
    Estimates phase phi such that U|psi> = e^{2pi i phi}|psi>.
    """
    n_anc = n_ancilla
    dim   = 2 ** n_anc
    # Build controlled-U^(2^k) phases for each ancilla
    eigs  = np.linalg.eigvals(unitary)
    # Pick eigenvalue closest to eigenstate
    evecs = np.linalg.eig(unitary)[1]
    overlaps = np.abs(evecs.T @ eigenstate)
    best_eig  = eigs[np.argmax(overlaps)]
    true_phase = np.angle(best_eig) / (2 * np.pi) % 1.0

    # Simulate QPE register via phase kickback
    register = np.zeros(dim, dtype=complex)
    register[0] = 1.0
    for k in range(n_anc):
        phase_k = (2 ** k) * true_phase
        register[1 << k] = np.exp(2j * np.pi * phase_k)
    register /= np.linalg.norm(register)

    # Inverse QFT
    iqft   = np.conj(_qft_matrix(n_anc)).T
    out    = iqft @ register
    probs  = np.abs(out) ** 2
    meas   = int(np.argmax(probs))
    est    = meas / dim

    return {
        "estimated_phase": est,
        "true_phase":      true_phase,
        "error":           abs(est - true_phase),
        "probabilities":   probs,
        "n_ancilla":       n_anc,
    }

def iterative_qpe(unitary: np.ndarray, eigenstate: np.ndarray, n_bits: int) -> dict:
    """Iterative QPE — one ancilla qubit, bit-by-bit phase readout."""
    eigs  = np.linalg.eigvals(unitary)
    evecs = np.linalg.eig(unitary)[1]
    overlaps  = np.abs(evecs.T @ eigenstate)
    best_eig  = eigs[np.argmax(overlaps)]
    true_phase = np.angle(best_eig) / (2 * np.pi) % 1.0

    bits   = []
    omega  = 0.0
    for k in range(n_bits - 1, -1, -1):
        phase_shift = -2 * np.pi * omega
        eff_phase   = (2 ** k) * true_phase + phase_shift / (2 * np.pi)
        prob_one    = np.sin(np.pi * eff_phase) ** 2
        bit         = 1 if np.random.random() < prob_one else 0
        bits.append(bit)
        omega       = (omega + bit * 2 ** (-(n_bits - k))) % 1.0

    estimated = sum(b * 2 ** (-(i + 1)) for i, b in enumerate(bits))
    return {
        "estimated_phase": estimated,
        "true_phase":      true_phase,
        "error":           abs(estimated - true_phase),
        "bits":            bits,
        "n_bits":          n_bits,
    }

def hamiltonian_qpe(hamiltonian: np.ndarray, eigenstate: np.ndarray,
                    n_ancilla: int, t: float = 1.0) -> dict:
    """HPE via matrix exponentiation: U = exp(-iHt)."""
    U = np.linalg.matrix_power(
        np.eye(hamiltonian.shape[0], dtype=complex),1)
    from scipy.linalg import expm
    U = expm(-1j * hamiltonian * t)
    result = qpe(U, eigenstate, n_ancilla)
    evals  = np.linalg.eigvalsh(hamiltonian)
    result["estimated_energy"] = result["estimated_phase"] * 2 * np.pi / t
    result["true_ground_energy"] = float(evals[0])
    return result




import numpy as np
import re

try:
    import cupy as cp
    HAS_GPU = True
except ImportError:
    cp = None
    HAS_GPU = False

class HPDensityMatrixResult:
    def __init__(self, counts, rho, n):
        self._counts = counts
        self._rho = rho
        self.num_qubits = n
    def get_counts(self): return self._counts
    def get_density_matrix(self): return self._rho
    def purity(self): return float(np.real(np.trace(self._rho @ self._rho)))
    def entropy(self):
        vals = np.linalg.eigvalsh(self._rho)
        vals = vals[vals > 1e-12]
        return float(-np.sum(vals * np.log2(vals)))
    def trace(self): return float(np.real(np.trace(self._rho)))
    def __repr__(self):
        return (f"HPDensityMatrixResult(qubits={self.num_qubits}, "
                f"purity={self.purity():.4f}, trace={self.trace():.4f})")


class HPDensityMatrixSimulator:
    _I    = np.eye(2, dtype=complex)
    _H    = np.array([[1,1],[1,-1]], dtype=complex) / np.sqrt(2)
    _X    = np.array([[0,1],[1,0]], dtype=complex)
    _Y    = np.array([[0,-1j],[1j,0]], dtype=complex)
    _Z    = np.array([[1,0],[0,-1]], dtype=complex)
    _S    = np.array([[1,0],[0,1j]], dtype=complex)
    _T    = np.array([[1,0],[0,np.exp(1j*np.pi/4)]], dtype=complex)
    _CNOT = np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]], dtype=complex)
    _CZ   = np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]], dtype=complex)
    _SWAP = np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]], dtype=complex)

    def __init__(self, use_gpu=False):
        self.use_gpu = use_gpu and HAS_GPU
        self.xp = cp if self.use_gpu else np

    def _to_xp(self, arr):
        return self.xp.array(arr) if self.use_gpu else np.array(arr)

    def _to_np(self, arr):
        return cp.asnumpy(arr) if (self.use_gpu and HAS_GPU) else np.array(arr)

    def _zero_rho(self, n):
        xp = self.xp
        rho = xp.zeros((2**n, 2**n), dtype=complex)
        rho[0, 0] = 1.0 + 0j
        return rho

    def _expand_1q(self, U, qubit, n):
        """Expand a 2x2 gate to the full 2^n x 2^n unitary."""
        full = None
        for i in range(n):
            m = U if i == qubit else self._I
            full = m if full is None else np.kron(full, m)
        return self._to_xp(full)

    def _expand_2q(self, U4, q0, q1, n):
        """Build full 2^n x 2^n unitary for a 2-qubit gate on q0, q1."""
        dim = 2**n
        result = np.zeros((dim, dim), dtype=complex)
        for col in range(dim):
            bits = [(col >> (n-1-i)) & 1 for i in range(n)]
            b0, b1 = bits[q0], bits[q1]
            amp_in = np.zeros(4, dtype=complex)
            amp_in[b0*2+b1] = 1.0
            amp_out = U4 @ amp_in
            for out_idx in range(4):
                ob0, ob1 = out_idx >> 1, out_idx & 1
                new_bits = bits.copy()
                new_bits[q0] = ob0; new_bits[q1] = ob1
                row = sum(new_bits[i] << (n-1-i) for i in range(n))
                result[row, col] += amp_out[out_idx]
        return self._to_xp(result)

    def _apply_unitary(self, rho, U):
        return U @ rho @ self.xp.conj(U).T

    def _apply_kraus(self, rho, kraus_list, qubit, n):
        """Apply single-qubit Kraus ops expanded to full system size."""
        xp = self.xp
        result = xp.zeros_like(rho)
        for K in kraus_list:
            # Expand 2x2 K to full 2^n space
            K_full = None
            for i in range(n):
                m = K if i == qubit else self._I
                K_full = m if K_full is None else np.kron(K_full, m)
            K_xp = self._to_xp(K_full)
            result = result + K_xp @ rho @ xp.conj(K_xp).T
        return result

    def _get_gate(self, name, params):
        if name == "H":  return self._H
        if name == "X":  return self._X
        if name == "Y":  return self._Y
        if name == "Z":  return self._Z
        if name == "S":  return self._S
        if name == "T":  return self._T
        if name == "RX":
            th = float(params[0]) if params else 0.0
            c,s = np.cos(th/2), np.sin(th/2)
            return np.array([[c,-1j*s],[-1j*s,c]], dtype=complex)
        if name == "RY":
            th = float(params[0]) if params else 0.0
            c,s = np.cos(th/2), np.sin(th/2)
            return np.array([[c,-s],[s,c]], dtype=complex)
        if name == "RZ":
            th = float(params[0]) if params else 0.0
            return np.array([[np.exp(-1j*th/2),0],[0,np.exp(1j*th/2)]], dtype=complex)
        return None

    def run(self, circuit, shots=1024, noise_model=None):
        n = circuit.num_qubits
        rho = self._zero_rho(n)

        for entry in circuit.gate_history:
            e = entry.lower()
            gate_name = None
            for g in ["rx","ry","rz","cnot","cz","swap","h","x","y","z","s","t"]:
                if e.startswith(g):
                    gate_name = g.upper(); break
            if gate_name is None: continue

            nums = list(map(int, re.findall(r"\d+", entry)))
            if not nums: continue
            floats = list(map(float, re.findall(r"[-+]?\d*\.\d+", entry)))

            if gate_name in ("CNOT","CZ","SWAP") and len(nums) >= 2:
                U4 = {"CNOT":self._CNOT,"CZ":self._CZ,"SWAP":self._SWAP}[gate_name]
                U_full = self._expand_2q(U4, nums[0], nums[1], n)
                rho = self._apply_unitary(rho, U_full)
            else:
                U1 = self._get_gate(gate_name, floats)
                if U1 is not None:
                    U_full = self._expand_1q(U1, nums[0], n)
                    rho = self._apply_unitary(rho, U_full)
                    if noise_model is not None:
                        err = noise_model.get_error(gate_name, nums[0])
                        if err:
                            # Pass qubit index and n so Kraus ops are expanded correctly
                            rho = self._apply_kraus(rho, err.kraus(), nums[0], n)

        rho_np = self._to_np(rho)
        probs = np.real(np.diag(rho_np))
        probs = np.clip(probs, 0, None)
        probs /= probs.sum()
        outcomes = np.random.choice(2**n, shots, p=probs)
        counts = {format(k, f"0{n}b"): int(v)
                  for k, v in zip(*np.unique(outcomes, return_counts=True))}
        return HPDensityMatrixResult(counts, rho_np, n)

    def partial_trace(self, rho_np, n, keep_qubits):
        trace_q = [q for q in range(n) if q not in keep_qubits]
        dm = rho_np.reshape([2]*(2*n))
        for q in sorted(trace_q, reverse=True):
            dm = np.trace(dm, axis1=q, axis2=q+len(dm.shape)//2)
        k = len(keep_qubits)
        return dm.reshape(2**k, 2**k)

    def properties(self):
        return {"name": "HPDensityMatrixSimulator",
                "backend": "CuPy-GPU" if self.use_gpu else "NumPy-CPU",
                "supports_noise": True, "supports_2qubit_gates": True,
                "max_qubits_recommended": 12}
